import unittest
import numpy as np
import scipy as sp
from scipy.io import mmread  # For importing files
import pathlib  # For importing files
import scipy.sparse as spsp
import scipy.sparse.linalg as spspla
import scipy
import pymoto as pym
import sys
import inspect

try:
    import cvxopt
    _has_cvxopt = True
except ImportError:
    sys.stdout.write("Skipping all tests with CVXOPT matrix...\n")
    _has_cvxopt = False

np.random.seed(0)  # Set seed for repeatability


def calculate_eigenvalues(A):
    """ Calculates the eigenvalues of a matrix to check positive-definiteness """
    return sp.linalg.eigvals(A.toarray())


def load_matrix(filen):
    """ Loads a matrix from file """
    curdir = pathlib.Path(__file__).parent.resolve()
    return mmread(pathlib.Path(curdir, "data", filen))


def generate_real_diagonal(N):
    """ Generates a real diagonal matrix of size N """
    return spsp.spdiags(np.random.rand(N), 0, N, N).tocoo()  # All positive


def generate_complex_diagonal(N):
    """ Generates a complex diagonal matrix of size N """
    return spsp.spdiags(np.random.rand(N) + 1j*np.random.rand(N), 0, N, N).tocoo()


def generate_symm_indef():
    """ Generates a real symmetric indefinite matrix """
    A = load_matrix("zenios.mtx.gz")
    sigma = -0.1  # To prevent exact singularity
    As = A - sigma*spsp.eye(A.shape[0])
    # Calculate eigenvalues
    # solver = pym.SolverSparseLU().update(As)
    # opinv = spspla.LinearOperator(A.shape, matvec=solver.solve, rmatvec=solver.adjoint)
    # w, _ = spspla.eigsh(A, sigma=sigma, k=100, OPinv=opinv, ncv=200)]
    return As.tocoo()


def generate_dynamic(complex=False, indef=False):
    """ Generate a dynamic stiffness matrix, either complex or not, and indefinite or not"""
    K = load_matrix("bcsstk26.mtx.gz")
    M = load_matrix("bcsstm26.mtx.gz")
    solver = pym.SolverSparseLU().update(K.tocsc())
    opinv = spspla.LinearOperator(K.shape, matvec=solver.solve, rmatvec=solver.adjoint)
    w, _ = spspla.eigsh(K, M=M, sigma=0, k=100, OPinv=opinv, ncv=200)
    eigfreqs = np.sqrt(w)
    if indef:
        freq = (eigfreqs[0] + eigfreqs[1])/2
    else:
        freq = eigfreqs[0]/2
    if complex:
        Ks = K + freq*1j*(0.01*K + 0.01*M) - freq**2 * M
    else:
        Ks = K - freq**2 * M
    return Ks.tocoo()


def generate_saddlepoint():
    K = load_matrix("bcsstk26.mtx.gz")
    M = load_matrix("bcsstm26.mtx.gz")
    solver = pym.SolverSparseLU().update(K.tocsc())
    opinv = spspla.LinearOperator(K.shape, matvec=solver.solve, rmatvec=solver.adjoint)
    w, V = spspla.eigsh(K, M=M, sigma=0, k=100, OPinv=opinv, ncv=200)
    wi, vi = w[0], V[:, [0]]
    Mv = M@vi

    return spsp.bmat([[K-wi*M, -Mv],
                      [-Mv.T, None]]).tocoo()


""" Perpare the set of matrices we want to test """
# Real-valued matrices
mat_real_spdiag = spsp.spdiags(np.random.rand(100), 0, 100, 100)  # Diagonal (spdiag format)
mat_real_diagonal = generate_real_diagonal(100)  # Diagonal (coo format)
mat_real_symm_pos_def = load_matrix("bcsstk14.mtx.gz")  # Symmetric positive definite
mat_real_symm_pos_def_dynamic = generate_dynamic(complex=False, indef=False)  # Symmetric positive definite
mat_real_symm_indef = generate_symm_indef()  # Symmetric indefinite
mat_real_symm_indef_dynamic = generate_dynamic(complex=False, indef=True)  # Symmetric indefinite
mat_real_symm_saddle = generate_saddlepoint()  # Indefinite saddlepoinnt
mat_real_asymm = load_matrix("impcol_a.mtx.gz")  # Asymmetric

# Complex-valued matrices
mat_complex_spdiag = mat_real_spdiag + 1j*mat_real_spdiag  # Diagonal (spdiag format)
mat_complex_diagonal = generate_complex_diagonal(100)  # Diagonal (coo format)
mat_complex_symm_pos_def_dynamic = generate_dynamic(complex=True, indef=False)  # Symmetric positive definite
mat_complex_symm_indef_dynamic = generate_dynamic(complex=True, indef=True)  # Symmetric indefinite
mat_complex_hermitian_1 = load_matrix("mhd1280b.mtx.gz")  # Hermitian
mat_complex_hermitian_indef = mat_complex_hermitian_1 - spsp.eye(*mat_complex_hermitian_1.shape)
mat_complex_hermitian_pos_def = mat_complex_hermitian_1 + spsp.eye(*mat_complex_hermitian_1.shape)

""" ------------------ TEST UTILITY FUNCTIONS -------------------- """


class TestGenericUtility(unittest.TestCase):
    """ Generic test class for testing functions on matrix properties that return True/False """
    data = []  # The data to test: list of tuples with (matrix, True/False property)
    fn = None  # The test function to use, which return the True/False property

    @classmethod
    def setUpClass(cls) -> None:
        """ Skip test if no function is given """
        if cls.fn is None:
            raise unittest.SkipTest(f"Skipping test {cls}")

    def get_tags(self):
        """ Generate tags for all the matrices to use in messaging"""
        src = inspect.getsource(type(self))
        start = src.find('data')
        start += src[start:].find('=') + 1
        finish = start + src[start:].find(']')

        def parse_str(t: str):
            rmv = ['\n', ' ', '(', ',', 'True', 'False']
            for c in rmv:
                t = t.replace(c, '')
            return t

        return [parse_str(t) for t in src[start:finish].strip(' []').split(')')]

    def test_all_matrices(self):
        """ Test all the matrices """
        tags = self.get_tags()
        for i, dat in enumerate(self.data):
            A = dat[0]
            res = dat[1]
            # Numpy dense matrix
            with self.subTest(msg=f"Dense - {tags[i]}"):
                # sys.stdout.write(f"Test {self.fn.__name__} for dense matrix {tags[i]}\n")
                self.assertEqual(self.fn(A.toarray()), res)

            # Scipy COO / spdiag matrix
            with self.subTest(msg=f"{type(A)} - {tags[i]}"):
                # sys.stdout.write(f"Test {self.fn.__name__} for {type(A)}-matrix {tags[i]}\n")
                self.assertEqual(self.fn(A), res)

            # Scipy CSR matrix
            with self.subTest(msg=f"CSR - {tags[i]}"):
                # sys.stdout.write(f"Test {self.fn.__name__} for CSR-matrix {tags[i]}\n")
                self.assertEqual(self.fn(A.tocsr()), res)

            # Scipy CSC matrix
            with self.subTest(msg=f"CSC - {tags[i]}"):
                # sys.stdout.write(f"Test {self.fn.__name__} for CSC-matrix {tags[i]}\n")
                self.assertEqual(self.fn(A.tocsc()), res)

            # CVXOPT spmatrix
            if _has_cvxopt:
                with self.subTest(msg=f"CVXOPT - {tags[i]}"):
                    # sys.stdout.write(f"Test {self.fn.__name__} for CVXOPT-matrix {tags[i]}\n")
                    if isinstance(A, spsp.coo_matrix):
                        Acvx = cvxopt.spmatrix(A.data, A.row.astype(int), A.col.astype(int))
                    self.assertEqual(self.fn(Acvx), res)


class TestIsComplex(TestGenericUtility):
    fn = staticmethod(pym.matrix_is_complex)
    data = [
        (mat_real_diagonal, False),
        (mat_real_spdiag, False),
        (mat_real_symm_pos_def, False),
        (mat_real_symm_pos_def_dynamic, False),
        (mat_real_symm_indef, False),
        (mat_real_symm_indef_dynamic, False),
        (mat_real_asymm, False),
        (mat_complex_diagonal, True),
        (mat_complex_spdiag, True),
        (mat_complex_symm_pos_def_dynamic, True),
        (mat_complex_symm_indef_dynamic, True),
        (mat_complex_hermitian_1, True),
        (mat_complex_hermitian_indef, True),
        (mat_complex_hermitian_pos_def, True),
    ]


class TestIsDiagonal(TestGenericUtility):
    fn = staticmethod(pym.matrix_is_diagonal)
    data = [
        (mat_real_diagonal, True),
        (mat_real_spdiag, True),
        (mat_real_symm_pos_def, False),
        (mat_real_symm_pos_def_dynamic, False),
        (mat_real_symm_indef, False),
        (mat_real_symm_indef_dynamic, False),
        (mat_real_asymm, False),
        (mat_complex_diagonal, True),
        (mat_complex_spdiag, True),
        (mat_complex_symm_pos_def_dynamic, False),
        (mat_complex_symm_indef_dynamic, False),
        (mat_complex_hermitian_1, False),
        (mat_complex_hermitian_indef, False),
        (mat_complex_hermitian_pos_def, False),
    ]


class TestIsSymmetric(TestGenericUtility):
    fn = staticmethod(pym.matrix_is_symmetric)
    data = [
        (mat_real_diagonal, True),
        (mat_real_spdiag, True),
        (mat_real_symm_pos_def, True),
        (mat_real_symm_pos_def_dynamic, True),
        (mat_real_symm_indef, True),
        (mat_real_symm_indef_dynamic, True),
        (mat_real_asymm, False),
        (mat_complex_diagonal, True),
        (mat_complex_spdiag, True),
        (mat_complex_symm_pos_def_dynamic, True),
        (mat_complex_symm_indef_dynamic, True),
        (mat_complex_hermitian_1, False),
        (mat_complex_hermitian_indef, False),
        (mat_complex_hermitian_pos_def, False),
    ]


class TestIsHermitian(TestGenericUtility):
    fn = staticmethod(pym.matrix_is_hermitian)
    data = [
        (mat_real_diagonal, True),
        (mat_real_spdiag, True),
        (mat_real_symm_pos_def, True),
        (mat_real_symm_pos_def_dynamic, True),
        (mat_real_symm_indef, True),
        (mat_real_symm_indef_dynamic, True),
        (mat_real_asymm, False),
        (mat_complex_diagonal, False),
        (mat_complex_spdiag, False),
        (mat_complex_symm_pos_def_dynamic, False),
        (mat_complex_symm_indef_dynamic, False),
        (mat_complex_hermitian_1, True),
        (mat_complex_hermitian_indef, True),
        (mat_complex_hermitian_pos_def, True),
    ]


""" ------------------ TEST THE SPARSE SOLVERS -------------------- """
# SOLVERS:
# - scipy/umfpack lu
# - scikit cholmod
# - cvxopt cholmod
# - intel mkl pardiso


class GenericTestSolvers(unittest.TestCase):
    """ Generic test runner for any solver and list of matrices """
    solver = None  # The solver to use
    matrices = []  # The list of matrices the solver should be able to handle

    @classmethod
    def setUpClass(cls) -> None:
        """ Skip test if the solver is not available """
        if cls.solver is None:
            raise unittest.SkipTest(f"No solver given for {cls.__name__}")
        if not cls.solver.defined:
            raise unittest.SkipTest(f"Solver \"{cls.solver.__name__}\" not available, probably lacking third-party libraries")

    def get_tags(self):
        """ Get the matrix tags for messages """
        src = inspect.getsource(type(self))
        start = src.find('matrices')
        start += src[start:].find('=') + 1
        finish = start + src[start:].find(']')
        return [t.strip(' ').replace('\n', '').replace(' ', '') for t in src[start:finish].strip(' []').split(',')]

    def run_solver(self, solver, A, b):
        """ Run the actual test for given solver, matrix and right-hand-side """
        # Reference solution
        x_ref = pym.SolverSparseLU(A).solve(b)
        xadj_ref = pym.SolverSparseLU(A.conj().T).solve(b)

        # Calculate the solution and the adjoint solution
        solver.update(A)
        x = solver.solve(b)
        xadj = solver.adjoint(b)

        # Check residual and solution
        residual = A @ x - b
        r = np.linalg.norm(residual) / np.linalg.norm(b)
        self.assertTrue(r < 1e-4, msg=f"Residual is {r}")
        self.assertTrue(np.allclose(x, x_ref))

        # Check adjoint residual and solution
        residual_adj = (A.conj().T) @ xadj - b
        r_adj = np.linalg.norm(residual_adj) / np.linalg.norm(b)
        self.assertTrue(r_adj < 1e-4, msg=f"Adjoint residual is {r_adj} (normal residual = {r})")
        self.assertTrue(np.allclose(xadj, xadj_ref))

    def run_sparse_tests(self, solver, Acoo, rhs):
        """ Run the sparse tests for different matrix types """
        self.run_solver(solver, Acoo, rhs)
        self.run_solver(solver, Acoo.tocsr(), rhs)
        self.run_solver(solver, Acoo.tocsc(), rhs)

    def test_all_matrices(self):
        """ Run the tests on all given matrices """
        tags = self.get_tags()
        for A, t in zip(self.matrices, tags):
            """ Test for different type of right-hand-sides """
            sys.stdout.write(f"Test \"{self.solver.__name__}\" for matrix \"{t}\"\n")
            N = A.shape[0]
            with self.subTest(msg=f"{t}.real-rhs"):
                b = np.random.rand(N)
                self.run_sparse_tests(self.solver(), A, b)
            with self.subTest(msg=f"{t}.multi-real-rhs"):
                b = np.random.rand(N, 3)
                self.run_sparse_tests(self.solver(), A, b)

            if not pym.matrix_is_complex(A):
                continue

            with self.subTest(msg=f"{t}.complex-rhs"):
                b = np.random.rand(N) + 1j * np.random.rand(N)
                self.run_sparse_tests(self.solver(), A, b)
            with self.subTest(msg=f"{t}.multi-complex-rhs"):
                b = np.random.rand(N, 3) + 1j * np.random.rand(N, 3)
                self.run_sparse_tests(self.solver(), A, b)


class TestSparseDiagonal(GenericTestSolvers):
    solver = pym.SolverDiagonal
    matrices = [
        mat_real_diagonal,
        mat_real_spdiag,
        mat_complex_diagonal,
        mat_complex_spdiag,
    ]


class TestSparseLU(GenericTestSolvers):
    solver = pym.SolverSparseLU
    matrices = [
        mat_real_diagonal,
        mat_real_spdiag,
        mat_real_symm_pos_def,
        mat_real_symm_pos_def_dynamic,
        mat_real_symm_indef,
        mat_real_symm_indef_dynamic,
        mat_real_asymm,
        mat_real_symm_saddle,
        mat_complex_diagonal,
        mat_complex_spdiag,
        mat_complex_symm_pos_def_dynamic,
        mat_complex_symm_indef_dynamic,
        mat_complex_hermitian_1,
        mat_complex_hermitian_indef,
        mat_complex_hermitian_pos_def,
    ]


class TestCholeskyScikit(GenericTestSolvers):
    # I don't know why, but scikit is able to solve indefinite matrix as well. Maybe they do some LDL inside?
    solver = pym.SolverSparseCholeskyScikit
    matrices = [
        mat_real_diagonal,
        mat_real_spdiag,
        mat_real_symm_pos_def,
        mat_real_symm_pos_def_dynamic,
        mat_complex_hermitian_indef,
        mat_complex_hermitian_1,
        mat_complex_hermitian_pos_def,
    ]


class TestCholeskyCVXOPT(GenericTestSolvers):
    solver = pym.SolverSparseCholeskyCVXOPT
    matrices = [
        mat_real_diagonal,
        mat_real_spdiag,
        mat_real_symm_pos_def,
        mat_real_symm_pos_def_dynamic,
        mat_complex_hermitian_1,
        mat_complex_hermitian_pos_def,
    ]


class TestPardiso(GenericTestSolvers):
    solver = pym.SolverSparsePardiso
    matrices = [
        mat_real_diagonal,
        mat_real_spdiag,
        mat_real_symm_pos_def,
        mat_real_symm_pos_def_dynamic,
        mat_real_symm_indef,
        mat_real_symm_indef_dynamic,
        mat_real_symm_saddle,
        mat_real_asymm,
        # mat_complex_diagonal, # PyPardiso does not work for complex matrix yet
        # mat_complex_spdiag,
        # mat_complex_symm_pos_def_dynamic,
        # mat_complex_symm_indef_dynamic,
        # mat_complex_hermitian_1,
    ]


""" ------------------ TEST LINSOLVE MODULE -------------------- """


class DynamicMatrix(pym.Module):
    alpha = 0.5
    beta = 0.5

    def _response(self, K, M, omega):
        return K + 1j * omega * (self.alpha * M + self.beta * K) - omega ** 2 * M

    def _sensitivity(self, dZ):
        K, M, omega = [s.state for s in self.sig_in]
        dK = np.real(dZ) - (omega * self.beta) * np.imag(dZ)
        dM = (-omega ** 2) * np.real(dZ) - (omega * self.alpha) * np.imag(dZ)
        dZrM = np.real(dZ).contract(M)
        dZiK = np.imag(dZ).contract(K)
        dZiM = np.imag(dZ).contract(M)
        domega = -self.beta * dZiK - self.alpha * dZiM - 2 * omega * dZrM
        return dK, dM, domega


class TestLinSolveModule_sparse(unittest.TestCase):
    # # ------------- Symmetric -------------
    def test_symmetric_real_compliance2d(self):
        """ Test symmetric real sparse matrix (compliance in 2D)"""
        N = 10  # Number of elements
        dom = pym.DomainDefinition(N, N)
        sx = pym.Signal('x', np.random.rand(dom.nel))
        fixed_nodes = dom.get_nodenumber(0, np.arange(0, N+1))
        bc = np.concatenate((fixed_nodes*2, fixed_nodes*2+1))
        iforce = dom.get_nodenumber(N, np.arange(0, N+1))*2 + 1
        sf = pym.Signal('f', np.zeros(dom.nnodes*2))
        sf.state[iforce] = 1.0

        fn = pym.Network()
        sK = fn.append(pym.AssembleStiffness(sx, pym.Signal('K'), dom, bc=bc))
        su = fn.append(pym.LinSolve([sK, sf], pym.Signal('u')))

        fn.response()

        self.assertTrue(np.allclose(sK.state@su.state, sf.state))  # Check residual
        # Check finite difference
        # def tfn(x0, dx, df_an, df_fd): np.allclose(df_an, df_fd, rtol=1e-3, atol=1e-5)
        def tfn(x0, dx, df_an, df_fd): self.assertTrue(np.allclose(df_an, df_fd, rtol=1e-3, atol=1e-5))
        pym.finite_difference(fn, [sx, sf], su, test_fn=tfn, dx=1e-5, tol=1e-4, verbose=False)

    def test_symmetric_real_compliance3d(self):
        """ Test symmetric real sparse matrix (compliance in 3D)"""
        N = 3  # Number of elements
        dom = pym.DomainDefinition(N, N, N)
        sx = pym.Signal('x', np.random.rand(dom.nel))
        jfix, kfix = np.meshgrid(np.arange(0, N+1), np.arange(0, N+1), indexing='ij')
        fixed_nodes = dom.get_nodenumber(0, jfix, kfix).flatten()
        bc = np.concatenate((fixed_nodes*3, fixed_nodes*3+1, fixed_nodes*3+2))
        iforce = dom.get_nodenumber(N, np.arange(0, N+1), np.arange(0, N+1))*3 + 1
        sf = pym.Signal('f', np.zeros(dom.nnodes*3))
        sf.state[iforce] = 1.0

        fn = pym.Network()
        sK = fn.append(pym.AssembleStiffness(sx, pym.Signal('K'), dom, bc=bc))
        su = fn.append(pym.LinSolve([sK, sf], pym.Signal('u')))

        fn.response()

        self.assertTrue(np.allclose(sK.state@su.state, sf.state))  # Check residual
        # Check finite difference
        # def tfn(x0, dx, df_an, df_fd): np.allclose(df_an, df_fd, rtol=1e-3, atol=1e-5)
        def tfn(x0, dx, df_an, df_fd): self.assertTrue(np.allclose(df_an, df_fd, rtol=1e-3, atol=1e-5))
        pym.finite_difference(fn, [sx, sf], su, test_fn=tfn, dx=1e-5, tol=1e-4, verbose=False)

    def test_symmetric_complex_dyncompliance2d(self):
        """ Test symmetric complex sparse matrix (dynamic compliance in 2D)"""
        N = 5  # Number of elements
        dom = pym.DomainDefinition(N, N)
        sx = pym.Signal('x', np.random.rand(dom.nel))
        fixed_nodes = dom.get_nodenumber(0, np.arange(0, N+1))
        bc = np.concatenate((fixed_nodes*2, fixed_nodes*2+1))
        iforce = dom.get_nodenumber(N, np.arange(0, N+1))*2 + 1
        sf = pym.Signal('f', np.zeros(dom.nnodes*2))
        sf.state[iforce] = 1.0

        sOmega = pym.Signal('omega', 0.1)
        fn = pym.Network()
        sK = fn.append(pym.AssembleStiffness(sx, pym.Signal('K'), dom, bc=bc))
        sM = fn.append(pym.AssembleMass(sx, pym.Signal('M'), dom, bc=bc))
        sZ = fn.append(DynamicMatrix([sK, sM, sOmega], pym.Signal('Z')))

        su = fn.append(pym.LinSolve([sZ, sf], pym.Signal('u')))

        fn.response()

        # spspla.eigsh(sK.state, M=sM.state, k=6, sigma=0.0)

        self.assertTrue(np.allclose(sZ.state@su.state, sf.state))  # Check residual
        # Check finite difference
        # def tfn(x0, dx, df_an, df_fd): np.allclose(df_an, df_fd, rtol=1e-3, atol=1e-5)
        def tfn(x0, dx, df_an, df_fd): self.assertTrue(np.allclose(df_an, df_fd, rtol=1e-3, atol=1e-5))
        pym.finite_difference(fn, [sx, sf, sOmega], su, test_fn=tfn, dx=1e-7, tol=1e-4, verbose=False)


if __name__ == '__main__':
    unittest.main()



import os
import platform
import numbers
import sys
from pathlib import Path
import numpy as np
if platform.system() == 'Darwin':  # Avoid "Python is not installed as a framework (Mac OS X)" error
    # Change backend
    import matplotlib
    matplotlib.use('TkAgg')
import matplotlib.pyplot as plt

from pymoto import Module
from .assembly import DomainDefinition


class _FigModule(Module):
    """ Abstract base class for any module which produces a figure

    Keyword Args:
        saveto (str): Save images of each iteration to the specified location. (default = ``None``)
        overwrite (bool): Overwrite saved image every time the figure is updated, else prefix ``_0000`` is added to the
          filename (default = ``False``)
        show (bool): Show the figure on the screen
    """
    def __init__(self, *args, saveto=None, overwrite=False, show=True, **kwargs):
        self.fig = plt.figure()
        if saveto is not None:
            self.saveloc, self.saveext = os.path.splitext(saveto)
            dir = os.path.dirname(saveto)
            if not os.path.exists(dir):  # Create dir
                os.makedirs(dir)
        else:
            self.saveloc, self.saveext = None, None
        self.overwrite = overwrite
        self.show = show
        self.iter = 0
        super().__init__(*args, **kwargs)

    def _update_fig(self):
        if self.iter == 0 and self.show:
            plt.show(block=False)

        self.fig.canvas.draw()
        self.fig.canvas.flush_events()

        if self.saveloc is not None:
            filen = "{0:s}{1:s}".format(self.saveloc, self.saveext) if self.overwrite else \
                "{0:s}_{1:04d}{2:s}".format(self.saveloc, self.iter, self.saveext)
            self.fig.savefig(filen)

        self.iter += 1

    def __del__(self):
        plt.close(self.fig)


class PlotDomain(_FigModule):
    """ Plots the densities of a domain (2D or 3D)

    Input Signal:
      - ``x`` (`np.ndarray`): The field to be shown of size ``(domain.nel)``

    Args:
        domain: The domain layout

    Keyword Args:
        saveto (str): Save images of each iteration to the specified location. (default = ``None``)
        overwrite (bool): Overwrite saved image every time the figure is updated, else prefix ``_0000`` is added to the
          filename (default = ``False``)
        show (bool): Show the figure on the screen
        clim: Color limits. In 2D ``[cmin, cmax]``: the values of minimum and maximum color. In 3D ``clipval``: the
          value below which elements are clipped.
        cmap (str): Colormap (only for 2D)
    """
    def _prepare(self, domain: DomainDefinition, clim=None, cmap='gray_r'):
        self.clim = clim
        self.cmap = cmap
        self.domain = domain

    def _response(self, x):
        if self.domain.dim == 2:
            self._plot_2d(x)
        elif self.domain.dim == 3:
            self._plot_3d(x)
        else:
            raise NotImplementedError("Only 2D and 3D plots are implemented")
        assert len(self.fig.axes) > 0, "Figure must contain axes"
        self.fig.axes[0].set_title(f"{self.sig_in[0].tag}, Iteration {self.iter}")

        self._update_fig()

    def _plot_2d(self, x):
        data = x.reshape((self.domain.nelx, self.domain.nely), order='F').T  # TODO Put this reshape inside domain?
        if hasattr(self, 'im'):
            self.im.set_data(data)
        else:
            ax = self.fig.add_subplot(111)
            self.im = ax.imshow(data, origin='lower', cmap=self.cmap)
            self.cbar = self.fig.colorbar(self.im, orientation='horizontal')
            ax.set(xlabel='x', ylabel='y')
        clim = [np.min(data), np.max(data)] if self.clim is None else self.clim
        self.im.set_clim(vmin=clim[0], vmax=clim[1])

    def _plot_3d(self, x):
        # prepare some coordinates, and attach rgb values to each
        ei, ej, ek = np.indices((self.domain.nelx, self.domain.nely, self.domain.nelz))
        els = self.domain.get_elemnumber(ei, ej, ek)
        densities = x[els]

        clip_lim = self.clim if self.clim is not None and isinstance(self.clim, numbers.Number) else 0.4

        sel = densities > clip_lim

        # combine the color components
        colors = np.zeros(sel.shape + (3,))
        colors[..., 0] = np.clip(1-densities, 0, 1)
        colors[..., 1] = np.clip(1-densities, 0, 1)
        colors[..., 2] = np.clip(1-densities, 0, 1)

        # and plot everything
        if len(self.fig.axes) == 0:
            # flake8: noqa: F401
            from mpl_toolkits.mplot3d import Axes3D  # This import is needed in order to support 3d plotting
            ax = self.fig.add_subplot(projection='3d')
            max_ext = max(self.domain.nelx, self.domain.nely, self.domain.nelz)
            ax.set(xlabel='x', ylabel='y', zlabel='z',
                   xlim=[(self.domain.nelx-max_ext)/2, (self.domain.nelx+max_ext)/2],
                   ylim=[(self.domain.nely-max_ext)/2, (self.domain.nely+max_ext)/2],
                   zlim=[(self.domain.nelz-max_ext)/2, (self.domain.nelz+max_ext)/2])
        else:
            ax = self.fig.axes[0]

        if hasattr(self, 'fac'):
            for i, f in self.fac.items():
                f.remove()

        self.fac = ax.voxels(sel, facecolors=colors, linewidth=0.5, edgecolors='k')


class PlotGraph(_FigModule):
    """ Plot an X-Y graph

    Input Signals:
      - ``x`` (`numpy.ndarray`): X-values
      - ``*args`` (`numpy.ndarray`): Y-values, which must match the dimension of ``x``

    Keyword Args:
        saveto (str): Save images of each iteration to the specified location. (default = ``None``)
        overwrite (bool): Overwrite saved image every time the figure is updated, else prefix ``_0000`` is added to the
          filename (default = ``False``)
        show (bool): Show the figure on the screen
        style (str): Line/marker style (*e.g.* ``"."``)
    """

    def _prepare(self, style: str = None):
        self.style = style

    def _response(self, x, *ys):
        if not hasattr(self, 'ax'):
            self.ax = self.fig.add_subplot(111)
            self.ax.set_xlabel(self.sig_in[0].tag)
            self.ax.set_ylabel(self.sig_in[1].tag)

        if not hasattr(self, 'line'):
            self.line = []
            for i, s in enumerate(self.sig_in[1:]):
                if self.style is None:
                    self.line.append(self.ax.plot([], [], label=s.tag)[0])
                else:
                    self.line.append(self.ax.plot([], [], self.style, label=s.tag)[0])
            self.ax.legend()

        ymin, ymax = np.inf, -np.inf
        for i, y in enumerate(ys):
            self.line[i].set_xdata(x)
            self.line[i].set_ydata(y)
            ymin, ymax = min(ymin, min(y)), max(ymax, max(y))
        self.ax.set_xlim([min(x), max(x)])
        self.ax.set_ylim([ymin, ymax])

        self._update_fig()


class PlotIter(_FigModule):
    """ Plot iteration history of one or more variables

    Input Signals:
      - ``*args`` (`Numeric` or `numpy.ndarray`): Y-values to show for each iteration

    Keyword Args:
        saveto (str): Save images of each iteration to the specified location. (default = ``None``)
        overwrite (bool): Overwrite saved image every time the figure is updated, else prefix ``_0000`` is added to the
          filename (default = ``False``)
        show (bool): Show the figure on the screen
    """
    def _prepare(self):
        self.minlim = 1e+200
        self.maxlim = -1e+200

    def _response(self, *args):
        if not hasattr(self, 'ax'):
            self.ax = self.fig.add_subplot(111)
            self.ax.set_yscale('linear')
            self.ax.set_xlabel("Iteration")

        if not hasattr(self, 'line'):
            self.line = []
            for i, s in enumerate(self.sig_in):
                self.line.append(None)
                self.line[i], = self.ax.plot([], [], '.', label=s.tag)
                self.ax.legend()

        for i, xx in enumerate(args):
            try:
                xadd = xx.reshape(xx.size)
                self.line[i].set_ydata(np.concatenate([self.line[i].get_ydata(), xadd]))
                self.line[i].set_xdata(np.concatenate([self.line[i].get_xdata(), self.iter*np.ones_like(xadd)]))
            except AttributeError:  # In case xx is not numpy, it doesn't have "reshape" nor "size" attributes
                xadd = xx
                self.line[i].set_ydata(np.append(self.line[i].get_ydata(), xadd))
                self.line[i].set_xdata(np.append(self.line[i].get_xdata(), self.iter))

            self.minlim = min(self.minlim, np.min(xadd))
            self.maxlim = max(self.maxlim, np.max(xadd))

        dy = max((self.maxlim - self.minlim)*0.05, sys.float_info.min)

        self.ax.set_xlim([-0.5, self.iter+0.5])
        if np.isfinite(self.minlim) and np.isfinite(self.maxlim):
            self.ax.set_ylim([self.minlim - dy, self.maxlim + dy])

        self._update_fig()


class WriteToVTI(Module):
    """ Writes vectors to a Paraview VTI file

    See also: :attr:`DomainDefinition.write_to_vti()`

    The size of the vectors should be a multiple of ``nel`` or ``nnodes``. Based on their size they are marked as
    cell-data or point-data in the VTI file. For 2D data (size is equal to ``2*nnodes``), the z-dimension is padded
    with zeros to have 3-dimensional data. Also block-vectors of multiple dimensions (*e.g.* ``(2, 3*nnodes)``) are
    accepted, which get the suffixed as ``_00``.

    Input Signals:
      - ``*args`` (`numpy.ndarary`): Vectors to write to VTI. The signal tags are used as name.

    Args:
        domain: The domain layout
        saveto (str): Location to save the VTI file
        overwrite (bool): Write a new file for each iteration
        scale (float): Scaling factor for the domain
    """
    def _prepare(self, domain: DomainDefinition, saveto: str, overwrite: bool = False, scale=1.):
        self.domain = domain
        self.saveto = saveto
        Path(saveto).parent.mkdir(parents=True, exist_ok=True)
        self.iter = 0
        self.scale = scale
        self.overwrite = overwrite

    def _response(self, *args):
        data = {}
        for s in self.sig_in:
            data[s.tag] = s.state
        pth = os.path.splitext(self.saveto)
        if self.overwrite:
            filen = pth[0] + pth[1]
        else:
            filen = pth[0] + '.{0:04d}'.format(self.iter) + pth[1]
        self.domain.write_to_vti(data, filename=filen, scale=self.scale)
        self.iter += 1



""" Specialized linear algebra modules """
import os
import sys
import glob
import warnings
import hashlib
from inspect import currentframe, getframeinfo

from pymoto import Module, DyadCarrier, LDAWrapper
import numpy as np
import scipy.sparse as sps
import scipy.linalg as spla  # Dense matrix solvers

from pymoto import SolverDenseLU, SolverDenseLDL, SolverDenseCholesky, SolverDiagonal, SolverDenseQR
from pymoto import matrix_is_symmetric, matrix_is_hermitian, matrix_is_diagonal
from pymoto import SolverSparseLU, SolverSparseCholeskyCVXOPT, SolverSparsePardiso, SolverSparseCholeskyScikit


class Inverse(Module):
    r""" Calculate the inverse of a matrix :math:`\mathbf{B} = \mathbf{A}^{-1}`

    Input Signal:
      - ``A`` (`np.ndarray`): Dense matrix:math:`\mathbf{A}`

    Output Signal:
      - `B` (`np.ndarray`): The inverse of :math:`\mathbf{A}` as dense matrix
    """
    def _response(self, A):
        return np.linalg.inv(A)

    def _sensitivity(self, dB):
        A = self.sig_in[0].state
        B = self.sig_out[0].state
        dA = - B.T @ dB @ B.T
        return dA if np.iscomplexobj(A) else np.real(dA)


# flake8: noqa: C901
def auto_determine_solver(A, isdiagonal=None, islowertriangular=None, isuppertriangular=None,
                          ishermitian=None, issymmetric=None, ispositivedefinite=None):
    """
    Uses parts of Matlab's scheme https://nl.mathworks.com/help/matlab/ref/mldivide.html
    :param A: The matrix
    :param isdiagonal: Manual override for diagonal matrix
    :param islowertriangular: Override for lower triangular matrix
    :param isuppertriangular: Override for upper triangular matrix
    :param ishermitian: Override for hermitian matrix (prevents check)
    :param issymmetric: Override for symmetric matrix (prevents check). Is the same as hermitian for a real matrix
    :param ispositivedefinite: Manual override for positive definiteness
    :return: LinearSolver which should be 'best' for the matrix
    """
    issparse = sps.issparse(A)  # Check if the matrix is sparse
    issquare = A.shape[0] == A.shape[1]  # Check if the matrix is square

    if not issquare:
        if issparse:
            sps.SparseEfficiencyWarning("Only a dense version of QR solver is available")  # TODO
        return SolverDenseQR()

    # l_bw, u_bw = spla.bandwidth(A) # TODO Get bandwidth (implemented in scipy version > 1.8.0)

    if isdiagonal is None:  # Check if matrix is diagonal
        # TODO: This could be improved to check other sparse matrix types as well
        isdiagonal = matrix_is_diagonal(A)
    if isdiagonal:
        return SolverDiagonal()

    # Check if the matrix is triangular
    # TODO Currently only for dense matrices
    if islowertriangular is None:  # Check if matrix is lower triangular
        islowertriangular = False if issparse else np.allclose(A, np.tril(A))
    if islowertriangular:
        warnings.WarningMessage("Lower triangular solver not implemented",
                                UserWarning, getframeinfo(currentframe()).filename, getframeinfo(currentframe()).lineno)

    if isuppertriangular is None:  # Check if matrix is upper triangular
        isuppertriangular = False if issparse else np.allclose(A, np.triu(A))
    if isuppertriangular:
        warnings.WarningMessage("Upper triangular solver not implemented",
                                UserWarning, getframeinfo(currentframe()).filename, getframeinfo(currentframe()).lineno)

    ispermutedtriangular = False
    if ispermutedtriangular:
        warnings.WarningMessage("Permuted triangular solver not implemented",
                                UserWarning, getframeinfo(currentframe()).filename, getframeinfo(currentframe()).lineno)

    # Check if the matrix is complex-valued
    iscomplex = np.iscomplexobj(A)
    if iscomplex:
        # Detect if the matrix is hermitian and/or symmetric
        if ishermitian is None:
            ishermitian = matrix_is_hermitian(A)
        if issymmetric is None:
            issymmetric = matrix_is_symmetric(A)
    else:
        if ishermitian is None and issymmetric is None:
            # Detect if the matrix is symmetric
            issymmetric = matrix_is_symmetric(A)
            ishermitian = issymmetric
        elif ishermitian is not None and issymmetric is not None:
            assert ishermitian == issymmetric, "For real-valued matrices, symmetry and hermitian must be equal"
        elif ishermitian is None:
            ishermitian = issymmetric
        elif issymmetric is None:
            issymmetric = ishermitian

    if issparse:
        # Prefer Intel Pardiso solver as it can solve any matrix TODO: Check for complex matrix
        if SolverSparsePardiso.defined and not iscomplex:
            # TODO check for positive definiteness?  np.alltrue(A.diagonal() > 0) or np.alltrue(A.diagonal() < 0)
            return SolverSparsePardiso(symmetric=issymmetric, hermitian=ishermitian, positive_definite=ispositivedefinite)

        if ishermitian:
            # Check if diagonal is all positive or all negative -> Cholesky
            if np.alltrue(A.diagonal() > 0) or np.alltrue(A.diagonal() < 0):  # TODO what about the complex case?
                if SolverSparseCholeskyScikit.defined:
                    return SolverSparseCholeskyScikit()
                if SolverSparseCholeskyCVXOPT.defined:
                    return SolverSparseCholeskyCVXOPT()

        return SolverSparseLU()  # Default to LU, which should be possible for any non-singular square matrix

    else:  # Dense branch
        if ishermitian:
            # Check if diagonal is all positive or all negative
            if np.alltrue(A.diagonal() > 0) or np.alltrue(A.diagonal() < 0):
                return SolverDenseCholesky()
            else:
                return SolverDenseLDL(hermitian=ishermitian)
        elif issymmetric:
            return SolverDenseLDL(hermitian=ishermitian)
        else:
            # TODO: Detect if the matrix is Hessenberg
            return SolverDenseLU()


class LinSolve(Module):
    r""" Solves linear system of equations :math:`\mathbf{A}\mathbf{x}=\mathbf{b}`

    Self-adjointness is automatically detected using :class:`LDAWrapper`.

    Input Signals:
      - ``A`` (`dense or sparse matrix`): The system matrix :math:`\mathbf{A}` of size ``(n, n)``
      - ``b`` (`vector`): Right-hand-side vector of size ``(n)`` or block-vector of size ``(Nrhs, n)``

    Output Signal:
      - ``x`` (`vector`): Solution vector of size ``(n)`` or block-vector of size ``(Nrhs, n)``

    Keyword Args:
        dep_tol: Tolerance for detecting linear dependence of solution vectors (default = ``1e-5``)
        hermitian: Flag to omit the automatic detection for Hermitian matrix, saves some work for large matrices
        symmetric: Flag to omit the automatic detection for symmetric matrix, saves some work for large matrices
        solver: Manually override the LinearSolver used, instead of the the solver from :func:`auto_determine_solver`

    Attributes:
        use_lda_solver: Use the linear-dependency-aware solver :class:`LDAWrapper` to prevent redundant computations
    """
    use_lda_solver = True

    def _prepare(self, dep_tol=1e-5, hermitian=None, symmetric=None, solver=None):
        self.dep_tol = dep_tol
        self.ishermitian = hermitian
        self.issymmetric = symmetric
        self.solver = solver

    def _response(self, mat, rhs):
        # Do some detections on the matrix type
        self.issparse = sps.issparse(mat)  # Check if it is a sparse matrix
        self.iscomplex = np.iscomplexobj(mat)  # Check if it is a complex-valued matrix
        if not self.iscomplex and self.issymmetric is not None:
            self.ishermitian = self.issymmetric
        if self.ishermitian is None:
            self.ishermitian = matrix_is_hermitian(mat)
        if self.issparse and not self.iscomplex and np.iscomplexobj(rhs):
            raise TypeError("Complex right-hand-side for a real-valued sparse matrix is not supported")

        # Determine the solver we want to use
        if self.solver is None:
            self.solver = auto_determine_solver(mat, ishermitian=self.ishermitian)
        if not isinstance(self.solver, LDAWrapper) and self.use_lda_solver:
            self.solver = LDAWrapper(self.solver, hermitian=self.ishermitian, symmetric=self.issymmetric)

        # Update solver with new matrix
        self.solver.update(mat)

        # Solution
        self.u = self.solver.solve(rhs)

        return self.u

    def _sensitivity(self, dfdv):
        mat, rhs = [s.state for s in self.sig_in]
        lam = self.solver.adjoint(dfdv.conj()).conj()

        if self.issparse:
            if not self.iscomplex and (np.iscomplexobj(self.u) or np.iscomplexobj(lam)):
                raise TypeError("Complex right-hand-side for a real-valued sparse matrix is not supported")  # TODO
                dmat = DyadCarrier([-np.real(lam), -np.imag(lam)], [np.real(self.u), np.imag(self.u)])
            else:
                dmat = DyadCarrier(-lam, self.u)
        else:
            if self.u.ndim > 1:
                dmat = np.einsum("iB,jB->ij", -lam, self.u, optimize=True)
            else:
                dmat = np.outer(-lam, self.u)
            if not self.iscomplex:
                dmat = np.real(dmat)

        db = np.real(lam) if np.isrealobj(rhs) else lam

        return dmat, db


class EigenSolve(Module):
    r""" Solves the (generalized) eigenvalue problem :math:`\mathbf{A}\mathbf{q}_i = \lambda_i \mathbf{B} \mathbf{q}_i`

    Solves the eigenvalue problem :math:`\mathbf{A}\mathbf{q}_i = \lambda_i \mathbf{q}_i` with normalization
    :math:`\mathbf{q}_i^\text{T} \mathbf{q}_i = 1` or
    the generalized eigenvalue problem :math:`\mathbf{A}\mathbf{q}_i = \lambda_i \mathbf{B} \mathbf{q}_i`
    with normalization :math:`\mathbf{q}_i^\text{T} \mathbf{B} \mathbf{q}_i = 1` (:math:`\mathbf{B}` must be positive
    definite). The eigenvectors are returned as a dense matrix ``Q``, where eigenvector ``Q[:, i]`` belongs to
    eigenvalue ``λ[i]``.

    Both real and complex matrices are supported.
    Correct sensitivities for cases with eigenvalue multiplicity are not implemented yet.

    Mode tracking algorithms can be implemented by the user by providing the argument ``sorting_func``, which is a
    function with arguments (``λ``, ``Q``).

    Todo:
        Support for sparse matrix

    Input Signal(s):
      - ``A`` (`dense matrix`): The system matrix of size ``(n, n)``
      - ``B`` (`dense matrix, optional`): Second system matrix (must be positive-definite) of size ``(n, n)``

    Output Signals:
      - ``λ`` (`vector`): Vector with eigenvalues of size ``(n)``
      - ``Q`` (`matrix`): Matrix with eigenvectors ``Q[:, i]`` corresponding to ``λ[i]``, of size ``(n, n)``

    Keyword Args:
        sorting_func: Sorting function to sort the eigenvalues, which must have signature ``func(λ,Q)``
          (default = ``numpy.argsort``)
        hermitian: Flag to omit the automatic detection for Hermitian matrix, saves some work for large matrices
    """
    def _prepare(self, sorting_func=lambda W, Q: np.argsort(W), hermitian=None):
        self.sorting_fn = sorting_func
        self.is_hermitian = hermitian

    def _response(self, A, *args):
        B = args[0] if len(args) > 0 else None
        if self.is_hermitian is None:
            self.is_hermitian = (matrix_is_hermitian(A) and (B is None or matrix_is_hermitian(B)))
        W, Q = spla.eigh(A, b=B) if self.is_hermitian else spla.eig(A, b=B)
        isort = self.sorting_fn(W, Q)
        W = W[isort]
        Q = Q[:, isort]
        for i in range(W.size):
            qi, wi = Q[:, i], W[i]
            qi *= np.sign(np.real(qi[0]))
            Bqi = qi if B is None else B@qi
            qi /= np.sqrt(qi@Bqi)
            assert (abs(qi@(qi if B is None else B@qi) - 1.0) < 1e-5)
            assert (np.linalg.norm(A@qi - wi*(qi if B is None else B@qi)) < 1e-5)
        return W, Q

    def _sensitivity(self, dW, dQ):
        A = self.sig_in[0].state
        B = self.sig_in[1].state if len(self.sig_in) > 1 else np.eye(*A.shape)
        dA = np.zeros_like(A)
        dB = np.zeros_like(B) if len(self.sig_in) > 1 else None
        W, Q = [s.state for s in self.sig_out]
        if dW is None:
            dW = np.zeros_like(W)
        if dQ is None:
            dQ = np.zeros_like(Q)

        for i in range(W.size):
            dqi, dwi = dQ[:, i], dW[i]
            if np.linalg.norm(dqi) == 0 and dwi == 0:
                continue
            qi, wi = Q[:, i], W[i]

            P = np.block([[(A - wi*B).T, -((B + B.T)/2)@qi[..., np.newaxis]], [-B@qi.T, 0]])
            adj = np.linalg.solve(P, np.block([dqi, dwi]))  # Adjoint solve Lee(1999)
            nu = adj[:-1]
            alpha = adj[-1]
            dAi = np.outer(-nu, qi)
            dA += dAi if np.iscomplexobj(A) else np.real(dAi)
            if dB is not None:
                dBi = np.outer(wi*nu + alpha/2*qi, qi)
                dB += dBi if np.iscomplexobj(B) else np.real(dBi)

        if len(self.sig_in) == 1:
            return dA
        elif len(self.sig_in) == 2:
            return dA, dB



import unittest
import pymoto as pym
import numpy as np


class TestSignal(unittest.TestCase):
    def test_initialize(self):
        a = pym.Signal('foo')
        self.assertEqual(a.tag, 'foo', msg="Initialize tag")
        self.assertIsNone(a.state, msg="Initialize state to None")
        self.assertIsNone(a.state, msg="Initialize sensitivity to None")

    def test_state(self):
        a = pym.Signal('foo')
        a.state = 1.0
        self.assertEqual(a.state, 1.0, msg="Set state to scalar")

        a.state = np.array([1.0, 2.0, 3.0])
        self.assertTrue(np.allclose(a.state, np.array([1.0, 2.0, 3.0])), msg="Set state to array")

        b = pym.Signal('foo', np.array([5.0, 6.0]))
        self.assertEqual(b.tag, 'foo', msg="Set tag from init with state")
        self.assertTrue(np.allclose(b.state, np.array([5.0, 6.0])), msg="Set state from init")

    def test_sensitivity(self):
        a = pym.Signal('foo')
        a.sensitivity = 2.0
        self.assertEqual(a.sensitivity, 2.0, msg="Set initial sensitivity to scalar")

        a.add_sensitivity(3.0)
        self.assertEqual(a.sensitivity, 5.0, msg="Add scalar sensitivity")

        a.sensitivity = 1.0
        self.assertEqual(a.sensitivity, 1.0, msg="Rewrite sensitivity by set_sens")

        a.reset(keep_alloc=True)
        self.assertEqual(a.sensitivity, 0.0, msg="Reset while keeping memory allocation")

        a.reset()
        self.assertEqual(a.sensitivity, None, msg="Reset sensitivity")

        a.add_sensitivity(np.array([1.1, 2.2, 3.3]))
        self.assertTrue(np.allclose(a.sensitivity, np.array([1.1, 2.2, 3.3])),
                        msg="Set initial sensitivity by add_sensitivity")

        a.add_sensitivity(None)
        self.assertTrue(np.allclose(a.sensitivity, np.array([1.1, 2.2, 3.3])),
                        msg="After adding None by add_sensitivity")

        b = pym.Signal('foo', np.array([5.0, 6.0]), np.array([7.0, 8.0]))
        self.assertEqual(b.tag, 'foo',
                         msg="Set tag from init with state and sensitivity")
        self.assertTrue(np.allclose(b.state, np.array([5.0, 6.0])),
                        msg="Set state from init and sensitivity")
        self.assertTrue(np.allclose(b.sensitivity, np.array([7.0, 8.0])),
                        msg="Set sensitivity from init and sensitivity")

        c = pym.Signal('bar', sensitivity=np.array([7.0, 8.0]))
        self.assertEqual(c.tag, 'bar', msg="Set tag from init with sensitivity")
        self.assertTrue(np.allclose(c.sensitivity, np.array([7.0, 8.0])),
                        msg="Set sensitivity from init with sensitivity")

    def test_make_signals(self):
        d = pym.make_signals('a', 'b', 'c')
        self.assertIsInstance(d['a'], pym.Signal)
        self.assertIsInstance(d['b'], pym.Signal)
        self.assertIsInstance(d['c'], pym.Signal)
        self.assertEqual(d['a'].tag, 'a')
        self.assertEqual(d['b'].tag, 'b')
        self.assertEqual(d['c'].tag, 'c')

    def test_add_sensitivity_errors(self):
        a = pym.Signal('foo')
        a.sensitivity = np.array([1.0, 2.0, 3.0])
        # Add wrong type
        self.assertRaises(TypeError, a.add_sensitivity, "cannot add a string")
        # a.add_sensitivity("cannot add a string")

        # Add wrong value
        self.assertRaises(ValueError, a.add_sensitivity, np.random.rand(3, 3))
        # a.add_sensitivity(np.random.rand(3, 3))

        # Adding complex array to real array does not give problems
        b = pym.Signal('real', np.random.rand(3))
        b.add_sensitivity(np.random.rand(3) + 1j*np.random.rand(3))

        c = pym.Signal('integer', 1)
        c.add_sensitivity(1.234)
        c.add_sensitivity(np.array(1.345))
        c.add_sensitivity(np.array([1.3344]))
        # c.add_sensitivity(np.array([[394]]))
        self.assertRaises(ValueError, c.add_sensitivity, np.array([[23454]]))  # Cannot add this shape to existing array

        d = pym.Signal('integer', 1.0)
        d.add_sensitivity(np.array([[23454]]))  # But it can be added to a float

        # Type which doesnt have +=
        class MyObj:
            def __init__(self, val):
                self.val = val
        e = pym.Signal('foo')
        e.add_sensitivity(MyObj(1.3))
        # e.add_sensitivity(MyObj(3.4))
        self.assertRaises(TypeError, e.add_sensitivity, MyObj(3.4))

    def test_reset_errors(self):
        a = pym.Signal('floatingpoint')
        a.add_sensitivity(1.3)
        a.reset(True)

        # With an object that doesnt have [] and *=
        class MyObj:
            def __init__(self, val):
                self.val = val
        b = pym.Signal('foo')
        b.add_sensitivity(MyObj(1.3)), self.assertWarns(RuntimeWarning, b.reset, True)
        # b.add_sensitivity(MyObj(1.3)), b.reset(True)  # Gives a warning, and just replaced by None


class TestSignalSlice(unittest.TestCase):
    def test_slice_1d(self):

        sx = pym.Signal('x')
        sx.state = np.random.rand(100)
        sx.state[1] = 0.314
        assert sx[0].state == sx.state[0]
        assert np.allclose(sx[0:5].state, sx.state[0:5])
        sx[0].state = 0.1
        assert sx[0].state == 0.1
        assert sx.state[0] == 0.1
        assert sx[1].state == 0.314

        sx[0].state += 0.1
        assert sx[0].state == 0.2
        assert sx.state[0] == 0.2

        sx[4:9].state = 0.1
        assert np.allclose(sx[4:9].state, 0.1)
        assert np.allclose(sx.state[4:9], 0.1)

        sx[4:9].state += 0.1
        assert np.allclose(sx[4:9].state, 0.2)
        assert np.allclose(sx.state[4:9], 0.2)

        assert sx[0].sensitivity is None
        assert sx[1:34, 53, 123:56, [2, 34, 5]].sensitivity is None

        sx[1].add_sensitivity(1.0)
        self.assertRaises(ValueError, sx[1].add_sensitivity, np.array([1, 2, 34]))
        assert sx.state.size == sx.sensitivity.size
        assert sx[1].state.size == sx[1].sensitivity.size
        assert sx[1].sensitivity == 1.0
        assert sx[0].sensitivity == 0.0

        sx[0].add_sensitivity(2.0)
        assert sx[0].sensitivity == 2.0
        assert sx.sensitivity[0] == 2.0
        assert sx[1].sensitivity == 1.0

        sx[0].reset()
        assert sx[0].sensitivity == 0.0
        assert sx[1].sensitivity == 1.0

        sx[4:8].add_sensitivity(3.0)
        assert sx[0].sensitivity == 0.0
        assert sx[1].sensitivity == 1.0
        assert np.allclose(sx[4:8].sensitivity, 3.0)
        assert np.allclose(sx.sensitivity[4:8], 3.0)
        assert np.allclose(sx[9].sensitivity, 0.0)

        sx[9:12].add_sensitivity(np.array([4, 5, 6]))
        self.assertRaises(ValueError, sx[9:12].add_sensitivity, np.array([4, 5, 6, 4]))
        assert np.allclose(sx[9:12].sensitivity, np.array([4, 5, 6]))
        sx[9:11].reset()
        assert np.allclose(sx[9:11].sensitivity, 0)
        assert sx[11].sensitivity == 6.0

    def test_slice_error(self):
        class MyObj:
            def __init__(self, val):
                self.val = val

        def call_state(s):
            return s.state

        def call_sens(s):
            return s.sensitivity

        # Object that cannot be sliced
        a = pym.Signal("myobj", state=MyObj(1.23), sensitivity=MyObj(1.34))
        # call_state(a[2]) # Empty state
        # call_sens(a[2])
        self.assertRaises(TypeError, call_state, a[2])
        self.assertRaises(TypeError, call_sens, a[2])

        # Too many dimensions
        b = pym.Signal("2dim", state=np.random.rand(10, 10), sensitivity=np.random.rand(10, 10))
        # call_state(b[2,3,4]) # Too many dimensions
        # call_sens(b[2,3,4])
        self.assertRaises(IndexError, call_state, b[2, 3, 4])
        self.assertRaises(IndexError, call_sens, b[2, 3, 4])
        # call_state(b[np.array([1,2,493]), 1]) # Out of range
        # call_sens(b[np.array([1,2,493]), 1])
        self.assertRaises(IndexError, call_state, b[np.array([1, 2, 493]), 1])
        self.assertRaises(IndexError, call_sens, b[np.array([1, 2, 493]), 1])

    def test_sensitivity_set_error(self):
        s = pym.Signal("empty")

        def set_sens(s, val):
            s.sensitivity = val

        # set_sens(s[2], 3.4) # Cannot set when state is None
        self.assertRaises(TypeError, set_sens, s[2], 3.4)

        class MyObj:
            def __init__(self, val):
                self.val = val
        a = pym.Signal("obj", MyObj(1.3))
        # set_sens(a[3], 3.4) # Cannot zero-initialize this object
        self.assertRaises(TypeError, set_sens, a[3], 3.4)

    def test_slice_2d(self):
        s = pym.Signal("2D_vals", np.random.rand(10, 10))

        s[0, 4].state = 0.4
        self.assertEqual(s[0, 4].state, 0.4)
        self.assertEqual(s.state[0, 4], 0.4)

        s[:, 3].state = 0.8
        self.assertTrue(np.alltrue(s[:, 3].state == 0.8))
        self.assertTrue(np.alltrue(s.state[:, 3] == 0.8))

        s[0, 2:8].sensitivity = 1.0
        self.assertEqual(s.sensitivity.shape, s.state.shape)
        self.assertTrue(np.alltrue(s.sensitivity[0, 2:8] == 1.0))
        self.assertEqual(s.sensitivity[0, 0], 0.0)
        self.assertEqual(s.state[0, 4], 0.4)  # Must be still the same as previously
        self.assertTrue(np.alltrue(s.state[:, 3] == 0.8))

        s[0, 2:8].sensitivity = 0.0
        self.assertTrue(np.alltrue(s.sensitivity == 0.0))

        # Test add_sensitivity
        add_arr = np.random.rand(6)
        s[0, 2:8].add_sensitivity(add_arr)
        self.assertTrue(np.alltrue(s.sensitivity[0, 2:8] == add_arr))
        s[0, 2:8].add_sensitivity(add_arr)
        self.assertTrue(np.alltrue(s.sensitivity[0, 2:8] == 2*add_arr))

        # Test reset
        s[0, 2:8].reset(keep_alloc=False)
        self.assertTrue(np.alltrue(s.sensitivity[0, 2:8] == 0))

        s[0, 2:8].add_sensitivity(add_arr)
        s[0, 2:8].reset(keep_alloc=True)
        self.assertTrue(np.alltrue(s.sensitivity[0, 2:8] == 0))

        s[0, 2:8].add_sensitivity(add_arr)
        s[0, 2:8].add_sensitivity(None)
        self.assertTrue(np.alltrue(s.sensitivity[0, 2:8] == add_arr))


class TestModule(unittest.TestCase):
    def test_initialize1(self):
        a = pym.Signal('x_in')
        b = pym.Signal('x_out')
        self.assertRaises(TypeError, pym.Module, a, b, msg="Can't instantiate the abstract base class without implementation")
        # pym.Module(a, b)

        class MyMod(pym.Module):
            def _response(self, x):
                return x*2

        mod = MyMod(a, b)
        self.assertEqual(len(mod.sig_in), 1, msg="Should have 1 input")
        self.assertEqual(len(mod.sig_out), 1, msg="Should have 1 output")
        self.assertEqual(mod.sig_in[0], a, msg="Input should be a")
        self.assertEqual(mod.sig_out[0], b, msg="Output should be b")

    def test_initialize2(self):
        a = pym.Signal('x1')
        b = pym.Signal('x2')
        c = pym.Signal('x3')
        d = pym.Signal('y1')
        e = pym.Signal('y2')

        class MyMod(pym.Module):
            def _response(self, x, y, z):
                return x+y, y+x

        bl = MyMod([a, b, c], [d, e])

        self.assertEqual(len(bl.sig_in), 3)
        self.assertEqual(len(bl.sig_out), 2)
        self.assertEqual(bl.sig_in[0], a)
        self.assertEqual(bl.sig_in[1], b)
        self.assertEqual(bl.sig_in[2], c)
        self.assertEqual(bl.sig_out[0], d)
        self.assertEqual(bl.sig_out[1], e)

    def test_create(self):
        class FooMod(pym.Module):
            def _response(self, a_in):
                return a_in * 2

        print("test_create: ")
        pym.Module.print_children()
        a = pym.Signal('x_in')
        b = pym.Signal('x_out')
        bl = pym.Module.create('foomod', a, b)

        self.assertIsInstance(bl, pym.Module)
        self.assertIsInstance(bl, FooMod)

        self.assertEqual(len(bl.sig_in), 1)
        self.assertEqual(len(bl.sig_out), 1)
        self.assertEqual(bl.sig_in[0], a)
        self.assertEqual(bl.sig_out[0], b)

        a.state = 1.0
        bl.response()
        self.assertEqual(b.state, 2.0)

        b.sensitivity = 1.0
        self.assertWarns(Warning, bl.sensitivity)
        # bl.sensitivity() # Warns about non-existent sensitivity function
        self.assertIsNone(a.sensitivity, msg="Default sensitivity behavior is None")

    def test_create_fail(self):
        self.assertRaises(ValueError, pym.Module.create, 'foomod1234', msg="Try to create a non-existing module")

        class FooMod(pym.Module):
            def _response(self, a_in):
                return a_in * 2

        a = pym.Signal('a')
        self.assertRaises(TypeError, FooMod, a, 1.0, msg="Try to initialize with invalid Signal object as output")
        # FooMod(a, 1.0)

        self.assertRaises(TypeError, FooMod, [1.0, 2], a, msg="Try initializing with invalid input Signal object")
        # FooMod([1.0, 2], a)

        b = pym.Signal('b')
        self.assertRaises(TypeError, FooMod, [b, 2], a, msg="Try initializing with invalid input Signal object")
        # FooMod([b, 2], a)

    def test_create_duplicate(self):
        class MathGeneral(pym.Module):
            def _response(self, a_in):
                return a_in * 2

        class Mathgeneral(pym.Module):
            def _response(self, a_in):
                return a_in * 2

        print(pym.Module.__subclasses__())
        self.assertWarns(Warning, pym.Module.print_children), "A warning should be emitted in case of duplicates"

        # Remove the duplicate module again
        del MathGeneral
        del Mathgeneral

        import gc
        gc.collect()

        print(pym.Module.__subclasses__())

        print("test_create_duplicate: ")
        pym.Module.print_children()

    def test_response_and_sens(self):
        class TwoInTwoOut(pym.Module):
            def _prepare(self, argument):
                self.prepared = argument

            def _response(self, a, b):
                self.internalstate = True
                return a * b, a + b

            def _sensitivity(self, dc, dd):
                self.didsensitivity = True
                a, b = [s.state for s in self.sig_in]
                return b * dc + dd, a * dc + dd

            def _reset(self):
                self.internalstate = False
                self.didsensitivity = False

        sa = pym.Signal('a', 2.5)
        sb = pym.Signal('b', 3.5)
        sc = pym.Signal('c')
        sd = pym.Signal('d')
        m = TwoInTwoOut([sa, sb], [sc, sd], 'foo')
        self.assertEqual(m.prepared, 'foo', msg="Check if the preparation has been executed")

        m.response()
        self.assertTrue(m.internalstate, msg="Check if response has been called")

        m.sensitivity()
        self.assertTrue(not hasattr(m, 'didsensitivity'), msg="Sensitivity should not have been called, "
                                                              "since output sensitivities are None")

        sc.sensitivity = 1.0
        sd.sensitivity = 1.0
        m.sensitivity()
        self.assertTrue(m.didsensitivity, msg="Sensitivity should have been called")

        self.assertEqual(sa.sensitivity, 4.5, msg="Check sensitivity value")
        self.assertEqual(sb.sensitivity, 3.5, msg="Check other sensitivity value")

        m.reset()
        self.assertFalse(m.internalstate, msg="Check if reset has worked")
        self.assertFalse(m.didsensitivity, msg="Check if reset has worked")

    def test_zero_inputs(self):
        class FooMod1(pym.Module):
            def _response(self):
                return 3.14

        b = pym.Signal('x_out')
        bl = FooMod1([], b)
        bl.response()
        self.assertEqual(b.state, 3.14)

    def test_zero_outputs(self):
        class FooMod2(pym.Module):
            def _response(self, in1):
                self.got_in1 = in1

            def _sensitivity(self):
                self.did_sens = True
                return 2.15

        a = pym.Signal('x_in', 1.256)
        bl = FooMod2(a)
        bl.response()
        self.assertEqual(bl.got_in1, 1.256, msg="State variable passed to _reponse function")

        bl.sensitivity()
        self.assertTrue(bl.did_sens, msg="Check if _sensitivity did run")
        self.assertEqual(a.sensitivity, 2.15, msg="After running first sensitivity")
        bl.sensitivity()
        self.assertEqual(a.sensitivity, 2.15 + 2.15, msg="After running second sensitivity")
        bl.reset()
        self.assertIsNone(a.sensitivity, msg="After resetting module")
        bl.sensitivity()
        self.assertEqual(a.sensitivity, 2.15, msg="First sensitivity run after reset")

    def test_wrong_response(self):
        class WrongResponse(pym.Module):
            """ Foobar

            """
            def _response(self, a):
                return a * 2.0, a * 3.0  # Two returns

        sa = pym.Signal('a', 2.5)
        sb = pym.Signal('b')
        m = WrongResponse(sa, sb)  # One output signal
        self.assertRaises(TypeError, m.response, msg="Number of out-signals should match number of returns in response")
        # m.response()

    def test_sensitivity_and_reset_errors(self):
        class NoSensitivity(pym.Module):
            def _response(self, a, b):
                return a * b

            def _sensitivity(self, dc):
                self.did_sensitivity = True
                b = self.sig_in[1].state
                return b * dc  # Only returns one sensitivity

            def _reset(self):
                raise RuntimeError("An error has occurred")

        sa = pym.Signal('a', 2.5)
        sb = pym.Signal('b', 3.5)
        sc = pym.Signal('c')
        m = NoSensitivity([sa, sb], sc)  # Two inputs -> expects two sensitivities returned

        m.response()
        m.sensitivity()  # First test with None as sensitivity
        self.assertTrue(not hasattr(m, "did_sensitivity"))

        sc.sensitivity = 1.0
        self.assertRaises(TypeError, m.sensitivity)
        # m.sensitivity()
        self.assertTrue(m.did_sensitivity)

        # m.reset()
        self.assertRaises(RuntimeError, m.reset)

        class ErrModule(pym.Module):
            def _response(self, a, b):
                return a * b

            def _sensitivity(self, dc):
                raise ValueError("some error in calculation")
        m1 = ErrModule([sa, sb], sc)
        m1.response()
        sc.sensitivity = 1.0
        self.assertRaises(ValueError, m1.sensitivity)
        # m1.sensitivity()


class TestNetwork(unittest.TestCase):
    def test_correct_network(self):
        x1 = pym.Signal('x1', 2.0)
        x2 = pym.Signal('x2', 3.0)
        y1 = pym.Signal('y1')
        y2 = pym.Signal('y2')
        z = pym.Signal('z')
        m1 = pym.MathGeneral(x1, y1, expression="x1*2.0")
        m2 = pym.MathGeneral(x2, y2, expression="x2*x2 + 2.0")
        m3 = pym.MathGeneral([y1, y2], z, expression="y1*y2")

        netw1 = pym.Network(m1, m2, m3)

        netw2 = pym.Network([m1, m2, {"type": "MathGeneral", "sig_in": [y1, y2], "sig_out": z,
                                      "expression": "y1*y2"}])  # Initalize with list

        netw1.response()
        self.assertEqual(y1.state, 4.0)
        self.assertEqual(y2.state, 11.0)
        self.assertEqual(z.state, 44.0)

        netw2.response()
        self.assertEqual(y1.state, 4.0)
        self.assertEqual(y2.state, 11.0)
        self.assertEqual(z.state, 44.0)

        z.sensitivity = 1.0
        netw1.sensitivity()
        self.assertEqual(y1.sensitivity, 11.0)
        self.assertEqual(y2.sensitivity, 4.0)
        self.assertEqual(x1.sensitivity, 22.0)
        self.assertEqual(x2.sensitivity, 24.0)

        netw1.reset()
        self.assertIsNone(x1.sensitivity)
        self.assertIsNone(x2.sensitivity)
        self.assertIsNone(y1.sensitivity)
        self.assertIsNone(y2.sensitivity)
        self.assertIsNone(z.sensitivity)

    def test_network_with_initializer_error(self):
        class ErrorModule(pym.Module):
            def _response(self, a1, a2):
                raise RuntimeError("Response error")

            def _sensitivity(self, dy):
                raise KeyError("Sensitivity error")

            def _reset(self):
                raise ValueError("Reset error")

        x1 = pym.Signal('x1', 2.0)
        x2 = pym.Signal('x2', 3.0)
        y1 = pym.Signal('y1')
        y2 = pym.Signal('y2')
        z = pym.Signal('z')
        m1 = pym.MathGeneral(x1, y1, expression="x1*2.0")
        m2 = pym.MathGeneral(x2, y2, expression="x2*x2 + 2.0")
        m3 = ErrorModule([y1, y2], z)

        self.assertRaises(KeyError, pym.Network, [m1, m2, {"sig_in": [y1, y2], "sig_out": z, "expression": "y1*y2"}])

        netw = pym.Network(m1, m2, m3)
        self.assertRaises(RuntimeError, netw.response)
        # netw.response()

        z.sensitivity = 1.0
        self.assertRaises(KeyError, netw.sensitivity)
        # netw.sensitivity()
        self.assertRaises(ValueError, netw.reset)
        # netw.reset()

        class PrepErrorModule(pym.Module):
            def _prepare(self):
                raise RuntimeError("Prepare error")

            def _response(self, a1, a2):
                raise RuntimeError("Response error")

        # pym.Network({'type': 'PrepErrorModule','sig_in': [], 'sig_out': []})
        self.assertRaises(RuntimeError, pym.Network, {'type': 'PrepErrorModule', 'sig_in': [], 'sig_out': []})


if __name__ == '__main__':
    unittest.main()



import unittest
import pymoto as pym
import numpy as np
import scipy.sparse as spsp


def generate_random(*dn, lower=-1.0, upper=1.0):
    return np.random.rand(*dn) * (upper - lower) + lower


class TestDyadCarrier(unittest.TestCase):
    # flake8: noqa: C901
    @staticmethod
    def setup_dyads(n=10, complex=False, nonsquare=False, empty=True, rnd=generate_random):
        dyads = {}

        if empty:
            dyads['empty'] = pym.DyadCarrier()

        ndyads = [1, 2, 10]
        for s in ndyads:
            dyads[f'square_{s}_dyads'] = pym.DyadCarrier([rnd(n) for _ in range(s)], [rnd(n) for _ in range(s)])

        if complex:
            for s in ndyads:
                dyads[f'square_complex_{s}_dyads'] = \
                    pym.DyadCarrier([rnd(n)+1j*rnd(n) for _ in range(s)],
                                    [rnd(n)+1j*rnd(n) for _ in range(s)])

        if nonsquare:
            nonsquare_offsets_u = [1]
            nonsquare_offsets_v = [2]

            for off_u in nonsquare_offsets_u:
                for s in ndyads:
                    dyads[f'nonsquare_u_{s}_dyads'] = \
                        pym.DyadCarrier([rnd(n+off_u) for _ in range(s)],
                                        [rnd(n) for _ in range(s)])
            for off_v in nonsquare_offsets_v:
                for s in ndyads:
                    dyads[f'nonsquare_v_{s}_dyads'] = \
                        pym.DyadCarrier([rnd(n) for _ in range(s)],
                                        [rnd(n+off_v) for _ in range(s)])

            if complex:
                for off_u in nonsquare_offsets_u:
                    for s in ndyads:
                        dyads[f'nonsquare_u_complex_{s}_dyads'] = \
                            pym.DyadCarrier([rnd(n+off_u)+1j*rnd(n+off_u) for _ in range(s)],
                                            [rnd(n)+1j*rnd(n) for _ in range(s)])
                for off_v in nonsquare_offsets_v:
                    for s in ndyads:
                        dyads[f'nonsquare_v_complex_{s}_dyads'] = \
                            pym.DyadCarrier([rnd(n)+1j*rnd(n) for _ in range(s)],
                                            [rnd(n+off_v)+1j*rnd(n+off_v) for _ in range(s)])
        return dyads

    def test_initialize(self):
        n = 10
        u1 = np.random.rand(n)
        u2 = np.random.rand(n)
        v1 = np.random.rand(n)
        v2 = np.random.rand(n)

        dyad1 = pym.DyadCarrier(u1, v1)
        self.assertEqual(len(dyad1.u), 1)
        self.assertEqual(len(dyad1.v), 1)
        self.assertEqual(len(dyad1.u[0]), len(u1))
        self.assertEqual(len(dyad1.v[0]), len(v1))
        self.assertTrue(np.allclose(dyad1.u[0], u1))
        self.assertTrue(np.allclose(dyad1.v[0], v1))
        self.assertEqual(dyad1.shape, (10, 10))

        dyad2 = pym.DyadCarrier([u1, u2], [v1, v2])
        self.assertEqual(len(dyad2.u), 2)
        self.assertEqual(len(dyad2.v), 2)
        self.assertEqual(len(dyad2.u[0]), len(u1))
        self.assertEqual(len(dyad2.u[1]), len(u2))
        self.assertEqual(len(dyad2.v[0]), len(v1))
        self.assertEqual(len(dyad2.v[1]), len(v2))
        self.assertTrue(np.allclose(dyad2.u[0], u1))
        self.assertTrue(np.allclose(dyad2.u[1], u2))
        self.assertTrue(np.allclose(dyad2.v[0], v1))
        self.assertTrue(np.allclose(dyad2.v[1], v2))
        self.assertEqual(dyad2.shape, (10, 10))

        dyad3 = pym.DyadCarrier(np.random.rand(2, n), np.random.rand(2, n))
        self.assertEqual(len(dyad3.u), 1)
        self.assertEqual(len(dyad3.v), 1)
        self.assertEqual(dyad3.shape, (10, 10))

        dyad4 = pym.DyadCarrier(np.random.rand(2, n), np.random.rand(3, n))
        self.assertEqual(len(dyad4.u), 1)
        self.assertEqual(len(dyad4.v), 1)
        self.assertEqual(dyad4.shape, (10, 10))

        ue = np.random.rand(n)
        ve = np.random.rand(2, n)
        dyad5 = pym.DyadCarrier(ue, ve)
        self.assertEqual(len(dyad5.u), 1)
        self.assertEqual(len(dyad5.v), 1)
        self.assertTrue(np.allclose(dyad5.u[0], ue))
        self.assertTrue(np.allclose(dyad5.v[0], ve[0, :] + ve[1, :]))
        self.assertEqual(dyad5.shape, (10, 10))

        uf = np.random.rand(5)
        vf = np.random.rand(10)
        dyad6 = pym.DyadCarrier(uf, vf)
        self.assertEqual(dyad6.shape, (5, 10))

        dyad7 = pym.DyadCarrier()
        self.assertEqual(len(dyad7.u), 0)
        self.assertEqual(len(dyad7.v), 0)
        dyad7.add_dyad(u1)
        self.assertEqual(len(dyad7.u), 1)
        self.assertEqual(len(dyad7.v), 1)
        self.assertTrue(np.allclose(dyad7.u[0], u1))
        self.assertTrue(np.allclose(dyad7.v[0], u1))
        dyad7.add_dyad(u2, v2)
        self.assertEqual(len(dyad7.u), 2)
        self.assertEqual(len(dyad7.v), 2)
        self.assertTrue(np.allclose(dyad7.u[1], u2))
        self.assertTrue(np.allclose(dyad7.v[1], v2))

        dyad8 = pym.DyadCarrier(u1)
        self.assertEqual(len(dyad8.u), 1)
        self.assertEqual(len(dyad8.v), 1)
        self.assertTrue(np.allclose(dyad8.u[0], u1))
        self.assertTrue(np.allclose(dyad8.v[0], u1))

        u9_1 = np.random.rand(2, 3, n)
        dyad9 = pym.DyadCarrier(u9_1, v1)
        self.assertEqual(len(dyad9.u), 1)
        self.assertEqual(len(dyad9.v), 1)
        uchk = u9_1[0, 0, :] + u9_1[0, 1, :] + u9_1[0, 2, :] + u9_1[1, 0, :] + u9_1[1, 1, :] + u9_1[1, 2, :]
        self.assertTrue(np.allclose(dyad9.u[0], uchk))
        self.assertTrue(np.allclose(dyad9.v[0], v1))

        self.assertRaises(TypeError, pym.DyadCarrier, u1, [v1, v2])

        self.assertRaises(TypeError, pym.DyadCarrier, [u1, np.random.rand(n + 1)], [v1, v2])

        self.assertRaises(TypeError, pym.DyadCarrier, [u1, [1.0, 2.0, 3.0]], [v1, v2])

        self.assertRaises(TypeError, pym.DyadCarrier, [u1, u2], [v1, np.random.rand(n + 1)])

        self.assertRaises(TypeError, pym.DyadCarrier, [u1, u2], [v1, [1.0, 2.0, 3.0]])

    def test_math_operations(self):
        n = 10
        u1 = np.random.rand(n)
        u2 = np.random.rand(n)
        v1 = np.random.rand(n)
        v2 = np.random.rand(n)

        a = pym.DyadCarrier(u1, v1)
        b = pym.DyadCarrier([u1, u2], [v1, v2])

        ap = +a
        self.assertFalse(a.u[0] is ap.u[0])
        self.assertFalse(a.v[0] is ap.v[0])
        self.assertTrue(np.allclose(a.u[0], ap.u[0]))
        self.assertTrue(np.allclose(a.v[0], ap.v[0]))

        bm = -b
        self.assertFalse(b.u[0] is bm.u[0])
        self.assertFalse(b.u[1] is bm.u[1])
        self.assertFalse(b.v[0] is bm.v[0])
        self.assertFalse(b.v[1] is bm.v[1])
        self.assertTrue(np.allclose(b.u[0], -bm.u[0]))
        self.assertTrue(np.allclose(b.u[1], -bm.u[1]))
        self.assertTrue(np.allclose(b.v[0], bm.v[0]))
        self.assertTrue(np.allclose(b.v[1], bm.v[1]))

        ap += b
        self.assertEqual(len(ap.u), 3)
        self.assertEqual(len(ap.v), 3)
        self.assertTrue(np.allclose(ap.u[0], a.u[0]))
        self.assertTrue(np.allclose(ap.v[0], a.v[0]))
        self.assertTrue(np.allclose(ap.u[1], b.u[0]))
        self.assertTrue(np.allclose(ap.v[1], b.v[0]))
        self.assertTrue(np.allclose(ap.u[2], b.u[1]))
        self.assertTrue(np.allclose(ap.v[2], b.v[1]))

        a1 = +a
        a1 -= b
        self.assertEqual(len(a1.u), 3)
        self.assertEqual(len(a1.v), 3)
        self.assertTrue(np.allclose(a1.u[0], a.u[0]))
        self.assertTrue(np.allclose(a1.v[0], a.v[0]))
        self.assertTrue(np.allclose(a1.u[1], -b.u[0]))
        self.assertTrue(np.allclose(a1.v[1], b.v[0]))
        self.assertTrue(np.allclose(a1.u[2], -b.u[1]))
        self.assertTrue(np.allclose(a1.v[2], b.v[1]))

        a2 = a + b
        self.assertEqual(len(a2.u), 3)
        self.assertEqual(len(a2.v), 3)
        self.assertTrue(np.allclose(a2.u[0], a.u[0]))
        self.assertTrue(np.allclose(a2.v[0], a.v[0]))
        self.assertTrue(np.allclose(a2.u[1], b.u[0]))
        self.assertTrue(np.allclose(a2.v[1], b.v[0]))
        self.assertTrue(np.allclose(a2.u[2], b.u[1]))
        self.assertTrue(np.allclose(a2.v[2], b.v[1]))

        a3 = a - b
        self.assertEqual(len(a3.u), 3)
        self.assertEqual(len(a3.v), 3)
        self.assertTrue(np.allclose(a3.u[0], a.u[0]))
        self.assertTrue(np.allclose(a3.v[0], a.v[0]))
        self.assertTrue(np.allclose(a3.u[1], -b.u[0]))
        self.assertTrue(np.allclose(a3.v[1], b.v[0]))
        self.assertTrue(np.allclose(a3.u[2], -b.u[1]))
        self.assertTrue(np.allclose(a3.v[2], b.v[1]))

        a4 = a.copy()
        self.assertFalse(a.u[0] is a4.u[0])
        self.assertFalse(a.v[0] is a4.v[0])
        self.assertTrue(np.allclose(a4.u[0], a.u[0]))
        self.assertTrue(np.allclose(a4.v[0], a.v[0]))

    def test_complex(self):
        n = 10
        u1 = np.random.rand(n)
        u2 = np.random.rand(n)
        v1 = np.random.rand(n)
        v2 = np.random.rand(n)
        u1c = u1 + 1j * np.random.rand(n)
        u2c = u2 + 1j * np.random.rand(n)
        v1c = v1 + 1j * np.random.rand(n)
        v2c = v2 + 1j * np.random.rand(n)

        a = pym.DyadCarrier([u1c, u2], [v1, v2])
        b = pym.DyadCarrier([u1, u2c], [v1, v2c])
        c = pym.DyadCarrier([u1c, u2c], [v1c, v2c])
        d = pym.DyadCarrier([u1, u2], [v1, v2])

        self.assertTrue(a.iscomplex(), msg="DyadCarrier should be recognized as complex, even if 1 entry is complex")
        self.assertTrue(b.iscomplex(), msg="DyadCarrier should be recognized as complex, even if 1 entry is complex")
        self.assertTrue(c.iscomplex(), msg="DyadCarrier should be recognized as complex")
        self.assertFalse(d.iscomplex(), msg="DyadCarrier should be recognized as real")

        # Complex conjugations
        self.assertTrue(np.allclose(np.conj(a.todense()), a.conj().todense()))
        self.assertTrue(np.allclose(np.conj(b.todense()), b.conj().todense()))

        cconj = c.conj()
        self.assertTrue(np.allclose(np.conj(c.u[0]), cconj.u[0]))
        self.assertTrue(np.allclose(np.conj(c.u[1]), cconj.u[1]))
        self.assertTrue(np.allclose(np.conj(c.v[0]), cconj.v[0]))
        self.assertTrue(np.allclose(np.conj(c.v[1]), cconj.v[1]))
        self.assertTrue(np.allclose(np.conj(c.todense()), cconj.todense()))

        dconj = d.conj()
        self.assertTrue(np.allclose(np.conj(d.u[0]), dconj.u[0]))
        self.assertTrue(np.allclose(np.conj(d.u[1]), dconj.u[1]))
        self.assertTrue(np.allclose(np.conj(d.v[0]), dconj.v[0]))
        self.assertTrue(np.allclose(np.conj(d.v[1]), dconj.v[1]))
        self.assertTrue(np.allclose(np.conj(d.todense()), dconj.todense()))

        # Get real part
        self.assertTrue(np.allclose(np.real(a.todense()), a.real.todense()))
        self.assertTrue(np.allclose(np.real(b.todense()), b.real.todense()))
        self.assertTrue(np.allclose(np.real(c.todense()), c.real.todense()))
        self.assertTrue(np.allclose(np.real(d.todense()), d.real.todense()))

        # Get imaginary part
        self.assertTrue(np.allclose(np.imag(a.todense()), a.imag.todense()))
        self.assertTrue(np.allclose(np.imag(b.todense()), b.imag.todense()))
        self.assertTrue(np.allclose(np.imag(c.todense()), c.imag.todense()))
        self.assertTrue(np.allclose(np.imag(d.todense()), d.imag.todense()))

    def test_todense(self):
        n = 10
        u1 = np.random.rand(n)
        u2 = np.random.rand(n)
        v1 = np.random.rand(n)
        v2 = np.random.rand(n)

        a = pym.DyadCarrier([u1, u2], [v1, v2])

        achk = np.outer(u1, v1) + np.outer(u2, v2)
        self.assertTrue(np.allclose(achk, a.todense()))

        uu1 = np.random.rand(2, n)
        vv1 = np.random.rand(2, n)
        b = pym.DyadCarrier(uu1, vv1)
        bchk = np.outer(uu1[0, :], vv1[0, :]) + np.outer(uu1[1, :], vv1[0, :]) \
            + np.outer(uu1[0, :], vv1[1, :]) + np.outer(uu1[1, :], vv1[1, :])
        self.assertTrue(np.allclose(b.todense(), bchk))

        cu1 = np.random.rand(2, 2, n)
        cv1 = np.random.rand(2, 2, n)
        c = pym.DyadCarrier(cu1, cv1)
        cchk = np.outer(cu1[0, 0, :], cv1[0, 0, :]) + np.outer(cu1[0, 1, :], cv1[0, 0, :]) \
            + np.outer(cu1[0, 0, :], cv1[0, 1, :]) + np.outer(cu1[0, 1, :], cv1[0, 1, :]) \
            + np.outer(cu1[1, 0, :], cv1[0, 0, :]) + np.outer(cu1[1, 1, :], cv1[0, 0, :]) \
            + np.outer(cu1[1, 0, :], cv1[0, 1, :]) + np.outer(cu1[1, 1, :], cv1[0, 1, :]) \
            + np.outer(cu1[0, 0, :], cv1[1, 0, :]) + np.outer(cu1[0, 1, :], cv1[1, 0, :]) \
            + np.outer(cu1[0, 0, :], cv1[1, 1, :]) + np.outer(cu1[0, 1, :], cv1[1, 1, :]) \
            + np.outer(cu1[1, 0, :], cv1[1, 0, :]) + np.outer(cu1[1, 1, :], cv1[1, 0, :]) \
            + np.outer(cu1[1, 0, :], cv1[1, 1, :]) + np.outer(cu1[1, 1, :], cv1[1, 1, :])
        self.assertTrue(np.allclose(c.todense(), cchk))

        d = pym.DyadCarrier([u1, uu1], [vv1, v1])
        dchk = np.outer(u1, vv1[0, :]) + np.outer(u1, vv1[1, :]) + np.outer(uu1[0, :], v1) + np.outer(uu1[1, :], v1)
        self.assertTrue(np.allclose(d.todense(), dchk))

        large_dyad = pym.DyadCarrier(np.random.rand(4000))
        self.assertWarns(ResourceWarning, large_dyad.todense)

        empty_dyad = pym.DyadCarrier()
        self.assertEqual(empty_dyad.todense().shape, (0, 0))

        u3 = np.random.rand(15)
        u4 = np.random.rand(15)

        dyad5 = pym.DyadCarrier([u3, u4], [v1, v2])
        dyad5chk = np.outer(u3, v1) + np.outer(u4, v2)
        self.assertTrue(np.allclose(dyad5chk, dyad5.todense()))

    def test_todense_complex(self):
        n = 10
        u1 = np.random.rand(n) + 1j * np.random.rand(n)
        u2 = np.random.rand(n)
        v1 = np.random.rand(n) + 1j * np.random.rand(n)
        v2 = np.random.rand(n) + 1j * np.random.rand(n)

        dyad1 = pym.DyadCarrier([u1, u2], [v1, v2])
        dyad1chk = np.outer(u1, v1) + np.outer(u2, v2)
        self.assertTrue(np.allclose(dyad1chk, dyad1.todense()))

        dyad2 = dyad1.conj()
        dyad2chk = np.conj(dyad1chk)
        self.assertTrue(np.allclose(dyad2chk, dyad2.todense()))

    def test_contract(self):
        n = 10
        u1 = np.random.rand(n)
        u2 = np.random.rand(n)
        v1 = np.random.rand(n)
        v2 = np.random.rand(n)

        a = pym.DyadCarrier([u1, u2], [v1, v2])

        tol = 1e-13
        self.assertAlmostEqual(a.contract(), np.dot(u1, v1) + np.dot(u2, v2), delta=tol)

        self.assertAlmostEqual(a.contract(np.eye(10)), np.dot(u1, v1) + np.dot(u2, v2), delta=tol)

        a_mat = np.random.rand(n, n)
        self.assertAlmostEqual(a.contract(a_mat), np.dot(u1, a_mat.dot(v1)) + np.dot(u2, a_mat.dot(v2)), delta=tol)

        a_submat = np.random.rand(3, 4)
        rows = np.array([3, 5, 5])
        cols = np.array([0, 1, 2, 8])

        self.assertAlmostEqual(a.contract(a_submat, rows, cols),
                               u1[rows].dot(a_submat.dot(v1[cols])) + u2[rows].dot(a_submat.dot(v2[cols])), delta=tol)
        self.assertAlmostEqual(a.contract(rows=cols, cols=cols),
                               np.dot(u1[cols], v1[cols]) + np.dot(u2[cols], v2[cols]), delta=tol)
        self.assertRaises(ValueError, a.contract, rows=rows, cols=cols)

        a_submat1 = np.random.rand(3, n)
        self.assertAlmostEqual(a.contract(a_submat1, rows),
                               u1[rows].dot(a_submat1.dot(v1)) + u2[rows].dot(a_submat1.dot(v2)), delta=tol)

        self.assertAlmostEqual(a.contract(a_submat1.T, cols=rows),
                               u1.dot(a_submat1.T.dot(v1[rows])) + u2.dot(a_submat1.T.dot(v2[rows])), delta=tol)

    def test_contract_batch(self):
        n = 10
        u1 = np.random.rand(n)
        u2 = np.random.rand(n)
        v1 = np.random.rand(n)
        v2 = np.random.rand(n)

        a = pym.DyadCarrier([u1, u2], [v1, v2])

        tol = 1e-13

        # Batch matrix
        a_mat = np.random.rand(3, n, n)
        self.assertEqual(a.contract(a_mat).shape, (3,))
        self.assertTrue(np.allclose(a.contract(a_mat), np.dot(u1, a_mat.dot(v1).T) + np.dot(u2, a_mat.dot(v2).T)))

        a_mat1 = np.random.rand(3, 4, n, n)
        self.assertEqual(a.contract(a_mat1).shape, (3, 4))
        self.assertTrue(np.allclose(a.contract(a_mat1), a_mat1.dot(v1).dot(u1) + a_mat1.dot(v2).dot(u2)))

        a_matfail = np.random.rand(n, n, 3, 4)
        self.assertRaises(ValueError, a.contract, a_matfail)

        # Batch rows
        a_submat = np.random.rand(3, 4)
        rows = np.array([[3, 5, 5], [5, 6, 7], [8, 9, 0]])
        cols = np.array([0, 1, 2, 8])
        self.assertEqual(a.contract(a_submat, rows, cols).shape, (3,))
        self.assertTrue(np.allclose(a.contract(a_submat, rows, cols),
                                    u1[rows].dot(a_submat.dot(v1[cols])) + u2[rows].dot(a_submat.dot(v2[cols]))))

        self.assertEqual(a.contract(rows=rows, cols=rows).shape, (3,))
        self.assertTrue(np.allclose(a.contract(rows=rows, cols=rows),
                                    np.sum(u1[rows] * v1[rows], axis=1) + np.sum(u2[rows] * v2[rows], axis=1)))

        cols_fail = np.array([[0, 1, 2, 8], [3, 4, 7, 8]])
        self.assertRaises(ValueError, a.contract, rows=rows, cols=cols_fail)

        # Batch cols
        a_submat1 = np.random.rand(3, 4)
        rows = np.array([3, 5, 5])
        cols = np.array([[0, 1, 2, 8], [3, 4, 7, 8], [5, 4, 1, 0]])
        self.assertEqual(a.contract(a_submat1, rows, cols).shape, (3,))
        self.assertTrue(np.allclose(a.contract(a_submat1, rows, cols),
                                    u1[rows].dot(a_submat1.dot(v1[cols].T)) + u2[rows].dot(a_submat1.dot(v2[cols].T))))

        # Batch all
        rows = np.array([[3, 5, 5], [5, 6, 7], [8, 9, 0]])
        cols = np.array([[0, 1, 2, 8], [3, 4, 7, 8], [5, 4, 1, 0]])
        a_submat2 = np.random.rand(3, 3, 4)
        self.assertAlmostEqual(a.contract(a_submat2, rows, cols)[0],
                               a.contract(a_submat2[0, :], rows[0, :], cols[0, :]), delta=tol)
        self.assertAlmostEqual(a.contract(a_submat2, rows, cols)[1],
                               a.contract(a_submat2[1, :], rows[1, :], cols[1, :]), delta=tol)
        self.assertAlmostEqual(a.contract(a_submat2, rows, cols)[2],
                               a.contract(a_submat2[2, :], rows[2, :], cols[2, :]), delta=tol)

        rows_fail = np.array([[3, 5, 5], [5, 6, 7]])
        self.assertRaises(ValueError, a.contract, a_submat2, rows_fail, cols)

    def test_contract_sparse(self):
        # Test contraction with a sparse matrix
        n = 10
        u1 = np.random.rand(n)
        u2 = np.random.rand(n)
        v1 = np.random.rand(n)
        v2 = np.random.rand(n)

        a = pym.DyadCarrier([u1, u2], [v1, v2])
        diag = np.random.rand(1, n)
        S = spsp.spdiags(diag, 0, m=diag.size, n=diag.size)  # Size must be given for scipy==1.7

        self.assertAlmostEqual(a.contract(S), np.sum(u1 * diag * v1 + u2 * diag * v2), delta=1e-10)

    def test_contract_sparse_slice(self):
        # Test contraction with a sliced sparse matrix
        n = 10
        u1 = np.random.rand(n)
        u2 = np.random.rand(n)
        v1 = np.random.rand(n)
        v2 = np.random.rand(n)

        a = pym.DyadCarrier([u1, u2], [v1, v2])
        diag = np.random.rand(1, n - 2)
        S = spsp.spdiags(diag, 0, m=diag.size, n=diag.size)

        self.assertAlmostEqual(a.contract(S, rows=np.arange(8), cols=np.arange(8)),
                               np.sum(u1[:8] * diag * v1[:8] + u2[:8] * diag * v2[:8]), delta=1e-10)

    def test_diagonal(self):
        n = 10
        dyads = self.setup_dyads(n, complex=True, nonsquare=True)
        k_list = np.arange(-12, 12)
        for k in k_list:
            for key, d in dyads.items():
                self.assertTrue(np.allclose(d.diagonal(k), np.diagonal(d.todense(), offset=k)),
                                msg=f"Failed diagonal test with offset \"{k}\" and dyad \"{key}\"")

    def test_add_to_zeroarray(self):
        n = 10
        zer = np.array(0, dtype=object)
        zer += pym.DyadCarrier(np.random.rand(n), np.random.rand(n))
        self.assertTrue(isinstance(zer, pym.DyadCarrier))

    def test_dot(self):
        n = 10
        dyads = self.setup_dyads(n, complex=True, nonsquare=True, empty=False)
        for key, d in dyads.items():
            q = generate_random(d.shape[1])
            chk = d.todense().dot(q)
            self.assertTrue(np.allclose(d.dot(q), chk), msg=f"Failed dot test with dyad \"{key}\"")
            # np.dot(d, q) # This form cannot be overridden, so it will result in a list of dyadcarriers

    def test_transpose(self):
        n = 10
        dyads = self.setup_dyads(n, complex=True, nonsquare=True, empty=False)
        for key, d in dyads.items():
            chk = d.todense().T
            res1 = d.T.todense()
            self.assertTrue(np.allclose(chk, res1), msg=f"Failed transpose test with dyad \"{key}\"")

    def test_matmul_vec(self):
        n = 10
        dyads = self.setup_dyads(n, complex=True, nonsquare=True, empty=False)
        for key, d in dyads.items():
            q = generate_random(d.shape[1])
            chk = d.todense() @ q
            res = d @ q
            self.assertTrue(np.allclose(chk, res), msg=f"Failed matmul with dyad \"{key}\"")

    def test_rmatmul_vec(self):
        n = 10
        dyads = self.setup_dyads(n, complex=True, nonsquare=True, empty=False)
        for key, d in dyads.items():
            q = generate_random(d.shape[0])
            chk = q @ (d.todense())
            res = q @ d
            self.assertTrue(np.allclose(chk, res), msg=f"Failed matmul with dyad \"{key}\"")

    def test_matmul(self):
        n = 10
        dyads = self.setup_dyads(n, complex=True, nonsquare=True, empty=False)
        for key, d in dyads.items():
            A = generate_random(d.shape[1], d.shape[1])
            chk = d.todense() @ A
            res = (d @ A).todense()
            self.assertTrue(np.allclose(chk, res), msg=f"Failed matmul with dyad \"{key}\"")

    def test_rmatmul(self):
        n = 10
        dyads = self.setup_dyads(n, complex=True, nonsquare=True, empty=False)
        for key, d in dyads.items():
            A = generate_random(d.shape[0], d.shape[0])
            chk = A @ (d.todense())
            res = (A @ d).todense()
            self.assertTrue(np.allclose(chk, res), msg=f"Failed matmul with dyad \"{key}\"")

    def test_matmul_dyad(self):
        n = 10
        dyads = self.setup_dyads(n, complex=True, nonsquare=True, empty=False)
        for key, d in dyads.items():
            A = d.todense()
            chk = A.T @ A
            res = (d.T @ d).todense()
            self.assertTrue(np.allclose(chk, res), msg=f"Failed matmul with 2 dyads \"{key}\"")

    def test_slice(self):
        n = 10
        dyads = self.setup_dyads(n, complex=True, nonsquare=True, empty=False)
        for key, d in dyads.items():
            A = d.todense()

            self.assertTrue(np.allclose(A[0:3, :], d[0:3, :].todense()))
            self.assertTrue(np.allclose(A[0, :], d[0, :]))
            self.assertTrue(np.allclose(A[:, 1], d[:, 1]))
            self.assertTrue(np.allclose(A[0:5, 1:3], d[0:5, 1:3].todense()))
            self.assertTrue(np.allclose(A[0, 1], d[0, 1]))
            self.assertTrue(np.allclose(A[3, 1], d[3, 1]))

            indi = np.arange(0, 4)

            self.assertTrue(np.allclose(A[indi, :], d[indi, :].todense()))
            self.assertTrue(np.allclose(A[:, indi], d[:, indi].todense()))
            self.assertTrue(np.allclose(A[indi, indi], d[indi, indi]))

            indi = np.array([1, 1, 2, 3, 3])
            self.assertTrue(np.allclose(A[indi, :], d[indi, :].todense()))
            self.assertTrue(np.allclose(A[:, indi], d[:, indi].todense()))
            self.assertTrue(np.allclose(A[indi, indi], d[indi, indi]))

            self.assertRaises(IndexError, lambda: d[np.array([1, 2]), np.array([1, 2, 3])])

    def test_slice_asignment(self):
        n = 10
        dyads = self.setup_dyads(n, complex=True, nonsquare=True, empty=False)
        for key, d in dyads.items():
            d0 = d.copy()
            A = d0.todense()
            A[0:3, :] = 0.0
            d[0:3, :] = 0.0
            self.assertTrue(np.allclose(A, d.todense()))

            A, d = d0.todense(), d0.copy()
            A[0, :] = 0.0
            d[0, :] = 0.0
            self.assertTrue(np.allclose(A, d.todense()))

            A, d = d0.todense(), d0.copy()
            A[:, 1] = 0.0
            d[:, 1] = 0.0
            self.assertTrue(np.allclose(A, d.todense()))

            # "Impossible" options for a low-rank matrix
            # A, d = d0.todense(), d0.copy()
            # A[0:5, 1:3] = 0.0
            # d[0:5, 1:3] = 0.0
            # self.assertTrue(np.allclose(A, d.todense()))

            # A, d = d0.todense(), d0.copy()
            # A[0, 1] = 0.0
            # d[0, 1] = 0.0
            # self.assertTrue(np.allclose(A, d.todense()))

            # A, d = d0.todense(), d0.copy()
            # A[3, 1] = 0.0
            # d[3, 1] = 0.0
            # self.assertTrue(np.allclose(A, d.todense()))

            indi = np.arange(0, 4)

            A, d = d0.todense(), d0.copy()
            A[indi, :] = 0.0
            d[indi, :] = 0.0
            self.assertTrue(np.allclose(A, d.todense()))

            A, d = d0.todense(), d0.copy()
            A[:, indi] = 0.0
            d[:, indi] = 0.0
            self.assertTrue(np.allclose(A, d.todense()))

            # self.assertTrue(np.allclose(A[indi, indi], d[indi, indi]))

            indi = np.array([1, 1, 2, 3, 3])
            A, d = d0.todense(), d0.copy()
            A[indi, :] = 0.0
            d[indi, :] = 0.0
            self.assertTrue(np.allclose(A, d.todense()))

            A, d = d0.todense(), d0.copy()
            A[:, indi] = 0.0
            d[:, indi] = 0.0
            self.assertTrue(np.allclose(A, d.todense()))
            # self.assertTrue(np.allclose(A[indi, indi], d[indi, indi]))

    def test_mul_with_scalar(self):
        n = 10
        dyads = self.setup_dyads(n, complex=True, nonsquare=True, empty=False)
        for key, d in dyads.items():
            A = d.todense()

            if np.iscomplexobj(A):
                c = generate_random(1) + 1j * generate_random(1)
            else:
                c = generate_random(1)

            chk = A * c
            res = (d * c).todense()
            self.assertTrue(np.allclose(chk, res))

            chk = A * c[0]
            res = (d * c[0]).todense()
            self.assertTrue(np.allclose(chk, res))

            chk = c * A
            res = (c * d).todense()
            self.assertTrue(np.allclose(chk, res))

            chk = c[0] * A
            res = (c[0] * d).todense()
            self.assertTrue(np.allclose(chk, res))


if __name__ == '__main__':
    unittest.main()



""" Assembly modules for finite element analysis """
import sys
from pymoto import Module, DyadCarrier, DomainDefinition
import numpy as np
from scipy.sparse import csc_matrix
from typing import Union
try:
    from opt_einsum import contract as einsum
except ModuleNotFoundError:
    from numpy import einsum


class AssembleGeneral(Module):
    r""" Assembles a sparse matrix according to element scaling :math:`\mathbf{A} = \sum_e x_e \mathbf{A}_e`

    Each element matrix is scaled and with the scaling parameter of that element
    :math:`\mathbf{A} = \sum_e x_e \mathbf{A}_e`.
    The number of degrees of freedom per node is deduced from the size of the element matrix passed into the module.
    For instance, in case an element matrix of shape ``(3*4, 3*4)`` gets passed with a 2D :class:`DomainDefinition`, the
    number of dofs per node equals ``3``.

    Input Signal:
        - ``x``: Scaling vector of size ``(Nel)``

    Output Signal:
        - ``A``: system matrix of size ``(n, n)``

    Args:
        domain: The domain-definition for which should be assembled
        element_matrix: The element matrix for one element :math:`\mathbf{K}_e`
        bc (optional): Indices of any dofs that are constrained to zero (Dirichlet boundary condition).
          These boundary conditions are enforced by setting the row and column of that dof to zero.
        bcdiagval (optional): Value to put on the diagonal of the matrix at dofs where boundary conditions are active.
        matrix_type (optional): The matrix type to construct. This is a constructor which must accept the arguments
          ``matrix_type((vals, (row_idx, col_idx)), shape=(n, n))``
    """
    def _prepare(self, domain: DomainDefinition, element_matrix: np.ndarray, bc=None, bcdiagval=None, matrix_type=csc_matrix):
        self.elmat = element_matrix
        self.ndof = self.elmat.shape[-1] // domain.elemnodes  # Number of dofs per node
        self.n = self.ndof * domain.nnodes  # Matrix size
        self.nnode = domain.nnodes

        self.dofconn = domain.get_dofconnectivity(self.ndof)

        # Row and column indices for the matrix
        self.rows = np.kron(self.dofconn, np.ones((domain.elemnodes*self.ndof, 1), dtype=int)).flatten()
        self.cols = np.kron(self.dofconn, np.ones((1, domain.elemnodes*self.ndof), dtype=int)).flatten()
        self.matrix_type = matrix_type

        # Boundary conditions
        self.bc = bc
        self.bcdiagval = np.max(element_matrix) if bcdiagval is None else bcdiagval
        if bc is not None:
            self.bcselect = np.argwhere(np.bitwise_not(np.bitwise_or(np.isin(self.rows, self.bc),
                                                                     np.isin(self.cols, self.bc)))).flatten()

            self.rows = np.concatenate((self.rows[self.bcselect], self.bc))
            self.cols = np.concatenate((self.cols[self.bcselect], self.bc))
        else:
            self.bcselect = None

    def _response(self, xscale: np.ndarray):
        scaled_el = ((self.elmat.flatten()[np.newaxis]).T * xscale).flatten(order='F')

        # Set boundary conditions
        if self.bc is not None:
            # Remove entries that correspond to bc before initializing
            mat_values = np.concatenate((scaled_el[self.bcselect], self.bcdiagval*np.ones(len(self.bc))))
        else:
            mat_values = scaled_el

        try:
            mat = self.matrix_type((mat_values, (self.rows, self.cols)), shape=(self.n, self.n))
        except TypeError as e:
            raise type(e)(str(e) + "\n\tInvalid matrix_type={}. Either scipy.sparse.cscmatrix or "
                                   "scipy.sparse.csrmatrix are supported"
                          .format(self.matrix_type)).with_traceback(sys.exc_info()[2]) from None

        return mat

    def _sensitivity(self, dgdmat: Union[DyadCarrier, np.ndarray]):
        if dgdmat.size <= 0:
            return [None]
        if self.bc is not None:
            dgdmat[self.bc, :] = 0.0
            dgdmat[:, self.bc] = 0.0
        if isinstance(dgdmat, np.ndarray):
            dx = np.zeros_like(self.sig_in[0].state)
            for i in range(len(dx)):
                indu, indv = np.meshgrid(self.dofconn[i], self.dofconn[i], indexing='ij')
                dx[i] = einsum("ij,ij->", self.elmat, dgdmat[indu, indv])
            return dx
        elif isinstance(dgdmat, DyadCarrier):
            return dgdmat.contract(self.elmat, self.dofconn, self.dofconn)


def get_B(dN_dx):
    """ Gets the strain-displacement relation (Cook, eq 3.1-9, P.80)

      - 1D : [ε_x]_i = B [u]_i
      - 2D : [ε_x; ε_y; γ_xy]_i = B [u, v]_i
      - 3D : [ε_x; ε_y; ε_z; γ_xy; γ_yz; γ_zx]_i = B [u, v, w]_i

    Args:
        dN_dx: Shape function derivatives [dNi_dxj] of size (#shapefn. x #dimensions)

    Returns:
        B strain-displacement relation of size (#strains x #shapefn.*#dimensions)
    """
    n_dim, n_shapefn = dN_dx.shape
    n_strains = int((n_dim * (n_dim+1))/2)  # Triangular number: ndim=3 -> nstrains = 3+2+1
    B = np.zeros((n_strains, n_shapefn*n_dim))
    if n_dim == 1:
        for i in range(n_shapefn):
            B[i, 0] = dN_dx[i, 0]
    elif n_dim == 2:
        for i in range(n_shapefn):
            B[:, i*n_dim:(i+1)*n_dim] = np.array([[dN_dx[0, i], 0],
                                                  [0,           dN_dx[1, i]],
                                                  [dN_dx[1, i], dN_dx[0, i]]])
    elif n_dim == 3:
        for i in range(n_shapefn):
            B[:, i*n_dim:(i+1)*n_dim] = np.array([[dN_dx[0, i], 0,           0],
                                                  [0,           dN_dx[1, i], 0],
                                                  [0,           0,           dN_dx[2, i]],
                                                  [dN_dx[1, i], dN_dx[0, i], 0],
                                                  [0,           dN_dx[2, i], dN_dx[1, i]],
                                                  [dN_dx[2, i], 0,           dN_dx[0, i]]])
    else:
        raise ValueError(f"Number of dimensions ({n_dim}) cannot be greater than 3")
    return B


def get_D(E: float, nu: float, mode: str = 'strain'):
    """ Get material constitutive relation for linear elasticity

    Args:
        E: Young's modulus
        nu: Poisson's ratio
        mode: Plane-``strain``, plane-``stress``, or ``3D``

    Returns:
        Material matrix
    """
    mu = E/(2*(1+nu))
    lam = (E*nu) / ((1+nu)*(1-2*nu))
    c1 = 2*mu + lam
    if 'strain' in mode.lower():
        return np.array([[c1,  lam, 0],
                         [lam, c1,  0],
                         [0,   0,   mu]])
    elif 'stress' in mode.lower():
        a = E / (1 - nu * nu)
        return a*np.array([[1,  nu, 0],
                           [nu, 1,  0],
                           [0,  0,  (1-nu)/2]])
    elif '3d' in mode.lower():
        return np.array([[c1,  lam, lam, 0,  0,  0],
                         [lam, c1,  lam, 0,  0,  0],
                         [lam, lam, c1,  0,  0,  0],
                         [0,   0,   0,   mu, 0,  0],
                         [0,   0,   0,   0,  mu, 0],
                         [0,   0,   0,   0,  0,  mu]])
    else:
        raise ValueError("Only for plane-stress, plane-strain, or 3d")


class AssembleStiffness(AssembleGeneral):
    r""" Stiffness matrix assembly by scaling elements in 2D or 3D
    :math:`\mathbf{K} = \sum_e x_e \mathbf{K}_e`

    Input Signal:
        - ``x``: Scaling vector of size ``(Nel)``

    Output Signal:
        - ``K``: Stiffness matrix of size ``(n, n)``

    Args:
        domain: The domain to assemble for -- this determines the element size and dimensionality
        args (optional): Other arguments are passed to AssembleGeneral

    Keyword Args:
        e_modulus: Young's modulus
        poisson_ratio: Poisson's ratio
        plane: Plane ``strain`` or plane ``stress``
        bcdiagval: The value to put on the diagonal in case of boundary conditions (bc)
        kwargs: Other keyword-arguments are passed to AssembleGeneral
    """
    def _prepare(self, domain: DomainDefinition, *args,
                 e_modulus: float = 1.0, poisson_ratio: float = 0.3, plane='strain', **kwargs):
        self.E, self.nu = e_modulus, poisson_ratio

        # Get material relation
        D = get_D(self.E, self.nu, '3d' if domain.dim == 3 else plane.lower())
        if domain.dim == 2:
            D *= domain.element_size[2]

        nnode = 2**domain.dim  # Number of nodes per element
        ndof = nnode*domain.dim

        # Element stiffness matrix
        self.KE = np.zeros((ndof, ndof))

        # Numerical integration
        siz = domain.element_size
        w = np.prod(siz[:domain.dim]/2)
        for n in domain.node_numbering:
            pos = n*(siz/2)/np.sqrt(3)  # Sampling point
            dN_dx = domain.eval_shape_fun_der(pos)
            B = get_B(dN_dx)
            self.KE += w * B.T @ D @ B  # Add contribution

        super()._prepare(domain, self.KE, *args, **kwargs)


class AssembleMass(AssembleGeneral):
    r""" Consistent mass matrix assembly by scaling elements
    :math:`\mathbf{M} = \sum_e x_e \mathbf{M}_e`

    Input Signal:
        - ``x``: Scaling vector of size ``(Nel)``

    Output Signal:
        - ``M``: Mass matrix of size ``(n, n)``

    Args:
        domain: The domain to assemble for -- this determines the element size and dimensionality
        *args: Other arguments are passed to AssembleGeneral

    Keyword Args:
        rho: Base density
        bcdiagval: The value to put on the diagonal in case of boundary conditions (bc)
        **kwargs : Other keyword-arguments are passed to AssembleGeneral
    """
    def _prepare(self, domain: DomainDefinition, *args, rho: float = 1.0, bcdiagval=0.0, **kwargs):
        # Element mass matrix
        # 1/36 Mass of one element
        mel = rho * np.prod(domain.element_size)

        # Consistent mass matrix
        ME = mel / 36 * np.array([[4.0, 0.0, 2.0, 0.0, 2.0, 0.0, 1.0, 0.0],
                                  [0.0, 4.0, 0.0, 2.0, 0.0, 2.0, 0.0, 1.0],
                                  [2.0, 0.0, 4.0, 0.0, 1.0, 0.0, 2.0, 0.0],
                                  [0.0, 2.0, 0.0, 4.0, 0.0, 1.0, 0.0, 2.0],
                                  [2.0, 0.0, 1.0, 0.0, 4.0, 0.0, 2.0, 0.0],
                                  [0.0, 2.0, 0.0, 1.0, 0.0, 4.0, 0.0, 2.0],
                                  [1.0, 0.0, 2.0, 0.0, 2.0, 0.0, 4.0, 0.0],
                                  [0.0, 1.0, 0.0, 2.0, 0.0, 2.0, 0.0, 4.0]])

        super()._prepare(domain, ME, *args, bcdiagval=bcdiagval, **kwargs)



r"""
Derivatives with respect to complex values are also supported by the pyMOTO framework. However, they behave a bit
differently than "normal" derivatives. A complex value (:math:`z = x+iy`) can be seen as having two independent
variables (*e.g.*  :math:`x` and :math:`y`, or, :math:`z` and its conjugate :math:`z^*`).

Definition of partial derivatives (Wirtinger derivatives) for the complex value :math:`z = x+iy` are defined as:

  - :math:`\frac{\partial}{\partial z} = \frac{1}{2}\left( \frac{\partial}{\partial x}
    - i \frac{\partial}{\partial y}\right)` (this is stored in :attr:`Signal.sensitivity`)
  - :math:`\frac{\partial}{\partial z^*} = \frac{1}{2}\left( \\frac{\partial}{\partial x}
    + i \frac{\partial}{\partial y}\right)`

Throughout the pyMOTO package, if :attr:`Signal.state` represents :math:`z`, the derivative with respect to any response
function :math:`f\in\mathbb{R}` that is stored in :attr:`Signal.sensitivity` *always* equals
:math:`\frac{\partial f}{\partial z}`. Note that this may be different in other packages that implement complex
derivatives (*e.g.*, `PyTorch <https://pytorch.org/docs/stable/notes/autograd.html#autograd-for-complex-numbers>`_,
JAX, or TensorFlow).

Useful identities (:math:`z\in\mathbb{C}` and :math:`s\in\mathbb{C}`):

  - :math:`\frac{\partial s^*}{\partial z^*} = \left( \frac{\partial s}{\partial z} \right)^*`
  - :math:`\frac{\partial s^*}{\partial z} = \left( \frac{\partial s}{\partial z^*} \right)^*`

And, in case :math:`f\in\mathbb{R}`:

  - :math:`\frac{\partial f}{\partial z^*} = \left( \frac{\partial f}{\partial z} \right)^*`

The chain rule for a mapping from :math:`z\in\mathbb{C}\rightarrow s\in\mathbb{C}\rightarrow f` can be seen as the
contributions of two independent variables (here, :math:`s` and :math:`s^*`):

:math:`\frac{\partial f}{\partial z} = \frac{\partial f}{\partial s}\frac{\partial s}{\partial z}
  + \frac{\partial f}{\partial s^*}\frac{\partial s^*}{\partial z}`

In case the intermediate variable is real, thus :math:`z\in\mathbb{C}\rightarrow r\in\mathbb{R}\rightarrow f`, the
chain rule reduces to

:math:`\frac{\partial f}{\partial z} = 2 \frac{\partial f}{\partial r}\frac{\partial r}{\partial z}`,

which may seem counter-intuitive, but compensates for the initial factor of 1/2.

For a mapping from real to complex, thus :math:`r\in\mathbb{R}\rightarrow z\in\mathbb{C}\rightarrow f`, the chain rule
becomes

:math:`\frac{\partial f}{\partial z} =
  2 \text{Re}\left( \frac{\partial f}{\partial s}\frac{\partial s}{\partial r} \right)`.

References:
  - `Wirtinger derivatives, Wikipedia <https://en.wikipedia.org/wiki/Wirtinger_derivatives>`_
  - Sarason (2007). *Complex function theory*. American Mathematical Society.
  - `Delgado (2009). *The complex gradient operator and the CR-calculus* <https://arxiv.org/pdf/0906.4835.pdf>`_
  - `Pytorch AutoGrad <https://pytorch.org/docs/stable/notes/autograd.html#autograd-for-complex-numbers>`_
"""

import numpy as np
from pymoto import Module


class MakeComplex(Module):
    """ Makes a complex variable from two real inputs :math:`z(x,y) = x + iy`

    Input Signals:
        - ``x``: Real component
        - ``y``: Imaginary component

    Output Signal:
        - ``z``: Complex value
    """
    def _response(self, x, y):
        return x + 1j*y

    def _sensitivity(self, dz):
        return np.real(dz), np.real(1j*dz)


class RealPart(Module):
    """ Takes the real part of a complex value :math:`x = \\text{Re}(z)`

    Input Signal:
        - ``z``: Complex value

    Output Signal:
        - ``x``: Real part
    """
    def _response(self, z):
        return np.real(z)

    def _sensitivity(self, dx):
        return np.real(dx)


class ImagPart(Module):
    """ Takes the imaginary part of a complex value :math:`y = \\text{Im}(z)`

    Input Signal:
        - ``z``: Complex value

    Output Signal:
        - ``y``: Imaginary part
    """
    def _response(self, z):
        return np.imag(z)

    def _sensitivity(self, dy):
        return -1j*dy


class ComplexNorm(Module):
    """ Takes the complex norm :math:`A = z z^*`

    Input Signal:
        - ``z``: Complex value

    Output Signal:
        - ``A``: Complex norm (real valued)
    """
    def _response(self, z):
        return np.absolute(z)

    def _sensitivity(self, dA):
        z = self.sig_in[0].state
        A = self.sig_out[0].state
        return 1/A*dA*np.conj(z)



import unittest
import numpy as np
import pymoto as pym


class TestMath(unittest.TestCase):
    def testVec_Scalar(self):
        """ Test functionality where a vector is multiplied with a scalar """
        sVec = pym.Signal("vec", np.array([1.0, 0.0, 3.8, 4.6]))
        sScalar = pym.Signal("scalar", 3.5)

        sRes = pym.Signal("result")

        mod = pym.MathGeneral([sVec, sScalar], sRes, expression="inp0*inp1")

        mod.response()

        # Check result
        self.assertIsInstance(sRes.state, type(sVec.state))
        self.assertTrue(np.allclose(sRes.state, np.array([1.0*3.5, 0.0*3.5, 3.8*3.5, 4.6*3.5])))

        sRes.sensitivity = np.array([1.0, -2.0, 0.0, -4.0])

        mod.sensitivity()

        # Check sensitivity types
        self.assertIsInstance(sScalar.sensitivity, type(sScalar.state))
        self.assertIsInstance(sVec.sensitivity, type(sVec.state))

        # Check results
        self.assertAlmostEqual(sScalar.sensitivity, 1.0*1.0 - 0.0*2.0 + 3.8*0.0 - 4.6*4.0)
        self.assertTrue(np.allclose(sVec.sensitivity, np.array([1.0*3.5, -2.0*3.5, 0.0*3.5, -4.0*3.5])))

    def testVec_npScalar(self):
        """ Test functionality where a vector is multiplied with a scalar """
        sVec = pym.Signal("vec", np.array([1.0, 0.0, 3.8, 4.6]))
        sScalar = pym.Signal("numpy_scalar", np.array(3.5))

        sRes = pym.Signal("result")

        mod = pym.MathGeneral([sVec, sScalar], sRes, expression="inp0*inp1")

        mod.response()

        # Check result
        self.assertIsInstance(sRes.state, type(sVec.state))
        self.assertTrue(np.allclose(sRes.state, np.array([1.0*3.5, 0.0*3.5, 3.8*3.5, 4.6*3.5])))

        sRes.sensitivity = np.array([1.0, -2.0, 0.0, -4.0])

        mod.sensitivity()

        # Check sensitivity types
        self.assertIsInstance(sScalar.sensitivity, type(sScalar.state))
        self.assertIsInstance(sVec.sensitivity, type(sVec.state))

        # Check results
        self.assertAlmostEqual(sScalar.sensitivity, 1.0*1.0 - 0.0*2.0 + 3.8*0.0 - 4.6*4.0)
        self.assertTrue(np.allclose(sVec.sensitivity, np.array([1.0*3.5, -2.0*3.5, 0.0*3.5, -4.0*3.5])))

    def testVec_scalar_complex(self):
        """ Test functionality where a vector is multiplied with a complex scalar """
        v = np.array([1.0, 0.0+2.1j, 3.8, 4.6+3.6j])
        s = 3.5 + 3.8j
        sVec = pym.Signal("vec", v.copy())
        sScalar = pym.Signal("scalar", s)

        sRes = pym.Signal("result")

        mod = pym.MathGeneral([sVec, sScalar], sRes, expression="inp0*inp1")

        mod.response()

        # Check result
        self.assertIsInstance(sRes.state, type(sVec.state))
        self.assertTrue(np.allclose(sRes.state, v*s))

        dr = np.array([1.0+0.1j, -2.0-2.5j, 0.0+0.8j, -4.0-2.2j])
        sRes.sensitivity = dr.copy()

        mod.sensitivity()

        # Check sensitivity types
        self.assertIsInstance(sScalar.sensitivity, type(sScalar.state))
        self.assertIsInstance(sVec.sensitivity, type(sVec.state))
        self.assertEqual(sVec.sensitivity.dtype, sVec.state.dtype)

        # Check results
        self.assertAlmostEqual(sScalar.sensitivity, dr@v)
        self.assertTrue(np.allclose(sVec.sensitivity, dr*s))

    def testVec_scalar_complex_real_input(self):
        """ Test functionality where a vector is multiplied with a complex scalar """
        v = np.array([1.0, 0.0 + 2.1j, 3.8, 4.6 + 3.6j])
        s = 3.5  # This is now a real input
        sVec = pym.Signal("vec", v.copy())
        sScalar = pym.Signal("scalar", s)

        sRes = pym.Signal("result")

        mod = pym.MathGeneral([sVec, sScalar], sRes, expression="inp0*inp1")

        mod.response()

        # Check result
        self.assertIsInstance(sRes.state, type(sVec.state))
        self.assertTrue(np.allclose(sRes.state, v * s))

        dr = np.array([1.0 + 0.1j, -2.0 - 2.5j, 0.0 + 0.8j, -4.0 - 2.2j])
        sRes.sensitivity = dr.copy()

        mod.sensitivity()

        # Check sensitivity types
        self.assertIsInstance(sScalar.sensitivity, type(sScalar.state))
        self.assertIsInstance(sVec.sensitivity, type(sVec.state))
        self.assertEqual(sVec.sensitivity.dtype, sVec.state.dtype)

        # Check results
        self.assertAlmostEqual(sScalar.sensitivity, np.real(dr @ v))
        self.assertTrue(np.allclose(sVec.sensitivity, dr * s))

    def testVec_npScalar_complex(self):
        """ Test functionality where a vector is multiplied with a scalar """
        v = np.array([1.0, 0.0 + 2.1j, 3.8, 4.6 + 3.6j])
        s = np.array(3.5 + 3.8j)
        sVec = pym.Signal("vec", v.copy())
        sScalar = pym.Signal("numpy_scalar", s.copy())

        sRes = pym.Signal("result")

        mod = pym.MathGeneral([sVec, sScalar], sRes, expression="inp0*inp1")

        mod.response()

        # Check result
        self.assertIsInstance(sRes.state, type(sVec.state))
        self.assertTrue(np.allclose(sRes.state, v*s))

        dr = np.array([1.0 + 0.1j, -2.0 - 2.5j, 0.0 + 0.8j, -4.0 - 2.2j])
        sRes.sensitivity = dr.copy()

        mod.sensitivity()

        # Check sensitivity types
        self.assertIsInstance(sScalar.sensitivity, type(sScalar.state))
        self.assertIsInstance(sVec.sensitivity, type(sVec.state))

        # Check results
        self.assertAlmostEqual(sScalar.sensitivity, dr@v)
        self.assertTrue(np.allclose(sVec.sensitivity, dr*s))

    def testVec_npScalar_complex_real_input(self):
        """ Test functionality where a vector is multiplied with a scalar """
        v = np.array([1.0, 0.0 + 2.1j, 3.8, 4.6 + 3.6j])
        s = np.array(3.5)
        sVec = pym.Signal("vec", v.copy())
        sScalar = pym.Signal("numpy_scalar", s.copy())

        sRes = pym.Signal("result")

        mod = pym.MathGeneral([sVec, sScalar], sRes, expression="inp0*inp1")

        mod.response()

        # Check result
        self.assertIsInstance(sRes.state, type(sVec.state))
        self.assertTrue(np.allclose(sRes.state, v*s))

        dr = np.array([1.0 + 0.1j, -2.0 - 2.5j, 0.0 + 0.8j, -4.0 - 2.2j])
        sRes.sensitivity = dr.copy()

        mod.sensitivity()

        # Check sensitivity types
        self.assertIsInstance(sScalar.sensitivity, type(sScalar.state))
        self.assertIsInstance(sVec.sensitivity, type(sVec.state))

        # Check results
        self.assertAlmostEqual(sScalar.sensitivity, np.real(v@dr))
        self.assertTrue(np.allclose(sVec.sensitivity, s*dr))

        def tfn(x0, dx, df_an, df_fd): self.assertTrue(np.allclose(df_an, df_fd, rtol=1e-7, atol=1e-5))
        pym.finite_difference(mod, test_fn=tfn)


if __name__ == '__main__':
    unittest.main()



__version__ = '1.0.1'

from .core_objects import Signal, Module, Network, make_signals
from .routines import finite_difference, minimize_oc, minimize_mma

# Imports from common
from .common.dyadcarrier import DyadCarrier
from .common.domain import DomainDefinition
from .common.solvers import matrix_is_complex, matrix_is_diagonal, matrix_is_symmetric, matrix_is_hermitian, \
    LinearSolver, LDAWrapper
from .common.solvers_dense import SolverDiagonal, SolverDenseQR, SolverDenseLU, SolverDenseCholesky, SolverDenseLDL
from .common.solvers_sparse import SolverSparsePardiso, SolverSparseLU, SolverSparseCholeskyScikit, SolverSparseCholeskyCVXOPT

# Import from modules
from .modules.generic import MathGeneral, EinSum
from .modules.linalg import Inverse, LinSolve, EigenSolve
from .modules.assembly import AssembleGeneral, AssembleStiffness, AssembleMass
from .modules.filter import FilterConv, Filter, DensityFilter, OverhangFilter
from .modules.io import PlotDomain, PlotGraph, PlotIter, WriteToVTI
from .modules.complex import MakeComplex, RealPart, ImagPart, ComplexNorm
from .modules.autodiff import AutoMod

__all__ = [
    'Signal', 'Module', 'Network', 'make_signals',
    'finite_difference', 'minimize_oc', 'minimize_mma',
    # Common
    'DyadCarrier',
    'DomainDefinition',
    'matrix_is_complex', 'matrix_is_diagonal', 'matrix_is_symmetric', 'matrix_is_hermitian',
    'LinearSolver', 'LDAWrapper',
    'SolverDiagonal', 'SolverDenseQR', 'SolverDenseLU', 'SolverDenseCholesky', 'SolverDenseLDL',
    'SolverSparsePardiso', 'SolverSparseLU', 'SolverSparseCholeskyScikit', 'SolverSparseCholeskyCVXOPT',
    # Modules
    "MathGeneral", "EinSum",
    "Inverse", "LinSolve", "EigenSolve",
    "AssembleGeneral", "AssembleStiffness", "AssembleMass",
    "FilterConv", "Filter", "DensityFilter", "OverhangFilter",
    "PlotDomain", "PlotGraph", "PlotIter", "WriteToVTI",
    "MakeComplex", "RealPart", "ImagPart", "ComplexNorm",
    "AutoMod"
]



import unittest
import numpy as np
import pymoto as pym


class TestAutoMod(unittest.TestCase):
    def test_automod_scalar(self):
        class Mult(pym.AutoMod):
            def _response(self, x, y):
                return x * y

        sx = pym.Signal("x", 1.1)
        sy = pym.Signal("y", 2.0)

        m = Mult([sx, sy], pym.Signal("z"))

        m.response()

        self.assertAlmostEqual(m.sig_out[0].state, 1.1*2.0)

        m.sig_out[0].sensitivity = 1.5

        m.sensitivity()

        self.assertAlmostEqual(sx.sensitivity, 1.5*2.0, 5)
        self.assertAlmostEqual(sy.sensitivity, 1.5*1.1, 5)

    def test_automod_vec(self):
        class MultSum(pym.AutoMod):
            def _response(self, x, y):
                return np.sum(x*y)

        sx = pym.Signal("x", np.array([1.1, 1.2, 1.3]))
        sy = pym.Signal("y", np.array([2.5, 2.6, 2.7]))

        m = MultSum([sx, sy], pym.Signal("z"))

        m.response()

        self.assertAlmostEqual(m.sig_out[0].state, 1.1*2.5 + 1.2*2.6 + 1.3*2.7)

        m.sig_out[0].sensitivity = 1.5

        m.sensitivity()

        self.assertTrue(np.allclose(sx.sensitivity, 1.5*sy.state))
        self.assertTrue(np.allclose(sy.sensitivity, 1.5*sx.state))


if __name__ == '__main__':
    unittest.main()
